# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Core constants module containing all constants."""

import string

FIELDS_SET = '__fields_set__'

ROOT = '__root__'

PACKAGE_OSCAL = 'trestle.oscal'

TRESTLE_CONFIG_DIR = '.trestle'
TRESTLE_DIST_DIR = 'dist'
TRESTLE_CONFIG_FILE = 'config.ini'
TRESTLE_KEEP_FILE = '.keep'
DRAWIO_FILE_EXT = '.drawio'
MARKDOWN_FILE_EXT = '.md'

ALLOWED_EXTENSIONS_IN_DIRS = {'.json', '.xml', '.yaml', '.yml', '.md'}

# these are hyphenated - no underscore - and mixed singular and plural
MODEL_TYPE_A_PLAN = 'assessment-plan'
MODEL_TYPE_A_RESULT = 'assessment-results'
MODEL_TYPE_CATALOG = 'catalog'
MODEL_TYPE_COMPDEF = 'component-definition'
MODEL_TYPE_POAM = 'plan-of-action-and-milestones'
MODEL_TYPE_PROFILE = 'profile'
MODEL_TYPE_SSP = 'system-security-plan'

# these are all plural and hyphenated - no underscore
MODEL_DIR_A_PLAN = 'assessment-plans'
MODEL_DIR_A_RESULT = 'assessment-results'
MODEL_DIR_CATALOG = 'catalogs'
MODEL_DIR_COMPDEF = 'component-definitions'
MODEL_DIR_POAM = 'plan-of-action-and-milestones'
MODEL_DIR_PROFILE = 'profiles'
MODEL_DIR_SSP = 'system-security-plans'

# these are singular except for assessment_results - and component, poam, and ssp are all abbreviated
# underscores here instead of hyphens
MODULE_NAME_A_PLAN = 'assessment_plan'
MODULE_NAME_A_RESULT = 'assessment_results'
MODULE_NAME_CATALOG = 'catalog'
MODULE_NAME_COMPDEF = 'component'
MODULE_NAME_POAM = 'poam'
MODULE_NAME_PROFILE = 'profile'
MODULE_NAME_SSP = 'ssp'

MODEL_MODULE_A_PLAN = f'{PACKAGE_OSCAL}.{MODULE_NAME_A_PLAN}'
MODEL_MODULE_A_RESULT = f'{PACKAGE_OSCAL}.{MODULE_NAME_A_RESULT}'
MODEL_MODULE_CATALOG = f'{PACKAGE_OSCAL}.{MODULE_NAME_CATALOG}'
MODEL_MODULE_COMPDEF = f'{PACKAGE_OSCAL}.{MODULE_NAME_COMPDEF}'
MODEL_MODULE_POAM = f'{PACKAGE_OSCAL}.{MODULE_NAME_POAM}'
MODEL_MODULE_PROFILE = f'{PACKAGE_OSCAL}.{MODULE_NAME_PROFILE}'
MODEL_MODULE_SSP = f'{PACKAGE_OSCAL}.{MODULE_NAME_SSP}'

MODEL_TYPE_LIST = [
    MODEL_TYPE_A_PLAN,
    MODEL_TYPE_A_RESULT,
    MODEL_TYPE_CATALOG,
    MODEL_TYPE_COMPDEF,
    MODEL_TYPE_POAM,
    MODEL_TYPE_PROFILE,
    MODEL_TYPE_SSP
]

MODEL_DIR_LIST = [
    MODEL_DIR_A_PLAN,
    MODEL_DIR_A_RESULT,
    MODEL_DIR_CATALOG,
    MODEL_DIR_COMPDEF,
    MODEL_DIR_POAM,
    MODEL_DIR_PROFILE,
    MODEL_DIR_SSP
]

MODEL_MODULE_LIST = [
    MODEL_MODULE_A_PLAN,
    MODEL_MODULE_A_RESULT,
    MODEL_MODULE_CATALOG,
    MODEL_MODULE_COMPDEF,
    MODEL_MODULE_POAM,
    MODEL_MODULE_PROFILE,
    MODEL_MODULE_SSP
]
"""Map of plural form of a model type to the oscal module that contains the classes related to it."""
MODEL_DIR_TO_MODEL_MODULE = {
    MODEL_DIR_A_PLAN: MODEL_MODULE_A_PLAN,
    MODEL_DIR_A_RESULT: MODEL_MODULE_A_RESULT,
    MODEL_DIR_CATALOG: MODEL_MODULE_CATALOG,
    MODEL_DIR_COMPDEF: MODEL_MODULE_COMPDEF,
    MODEL_DIR_POAM: MODEL_MODULE_POAM,
    MODEL_DIR_PROFILE: MODEL_MODULE_PROFILE,
    MODEL_DIR_SSP: MODEL_MODULE_SSP
}
"""Map of model type to oscal module."""
MODEL_TYPE_TO_MODEL_MODULE = {
    MODEL_TYPE_A_PLAN: MODEL_MODULE_A_PLAN,
    MODEL_TYPE_A_RESULT: MODEL_MODULE_A_RESULT,
    MODEL_TYPE_CATALOG: MODEL_MODULE_CATALOG,
    MODEL_TYPE_COMPDEF: MODEL_MODULE_COMPDEF,
    MODEL_TYPE_POAM: MODEL_MODULE_POAM,
    MODEL_TYPE_PROFILE: MODEL_MODULE_PROFILE,
    MODEL_TYPE_SSP: MODEL_MODULE_SSP
}
"""Map of model module to model type."""
MODEL_MODULE_TO_MODEL_TYPE = {
    MODEL_MODULE_A_PLAN: MODEL_TYPE_A_PLAN,
    MODEL_MODULE_A_RESULT: MODEL_TYPE_A_RESULT,
    MODEL_MODULE_CATALOG: MODEL_TYPE_CATALOG,
    MODEL_MODULE_COMPDEF: MODEL_TYPE_COMPDEF,
    MODEL_MODULE_POAM: MODEL_TYPE_POAM,
    MODEL_MODULE_PROFILE: MODEL_TYPE_PROFILE,
    MODEL_MODULE_SSP: MODEL_TYPE_SSP
}
"""Map of model type to model directory."""
MODEL_TYPE_TO_MODEL_DIR = {
    MODEL_TYPE_A_PLAN: MODEL_DIR_A_PLAN,
    MODEL_TYPE_A_RESULT: MODEL_DIR_A_RESULT,
    MODEL_TYPE_CATALOG: MODEL_DIR_CATALOG,
    MODEL_TYPE_COMPDEF: MODEL_DIR_COMPDEF,
    MODEL_TYPE_POAM: MODEL_DIR_POAM,
    MODEL_TYPE_PROFILE: MODEL_DIR_PROFILE,
    MODEL_TYPE_SSP: MODEL_DIR_SSP
}
"""Element path separator"""
ALIAS_PATH_SEPARATOR: str = '.'

# argument names
ARG_FILE = 'file'
ARG_FILE_SHORT = 'f'

ARG_ELEMENT = 'element'
ARG_ELEMENT_SHORT = 'e'

# argument descriptions
ARG_DESC_FILE = 'Path of the file'
ARG_DESC_ELEMENT = 'Path of the element in the OSCAL model'

# Index separater for naming directories representing collection properties
IDX_SEP = '__'

# Digit prefix for split files
FILE_DIGIT_PREFIX_LENGTH = 5

# Wildcard that can be used in the element path to represent all elements
ELEMENT_WILDCARD = '*'
ARG_VALIDATE = 'validate'
ARG_VALIDATE_SHORT = 'v'

ARG_MODE = 'mode'
ARG_MODE_SHORT = 'm'
ARG_DESC_MODE = 'Mode of the operation'

ITEM = 'item'

VAL_MODE_CATALOG = 'catalog'
VAL_MODE_DUPLICATES = 'duplicates'
VAL_MODE_LINKS = 'links'
VAL_MODE_REFS = 'refs'
VAL_MODE_RULES = 'rules'
VAL_MODE_ALL = 'all'

IOF_SHORT = '-iof'
IOF_LONG = '--include-optional-fields'
IOF_HELP = 'Include fields that are optional in the OSCAL model when generating the new object.'

INIT_FULL_SHORT = '-fl'
INIT_FULL_LONG = '--full'
INIT_FULL_HELP = 'Initializes Trestle workspace for local, API and governed documents usage.'

INIT_GOVDOCS_SHORT = '-gd'
INIT_GOVDOCS_LONG = '--govdocs'
INIT_GOVDOCS_HELP = 'Initializes Trestle workspace for governed documents usage only.'

INIT_LOCAL_SHORT = '-loc'
INIT_LOCAL_LONG = '--local'
INIT_LOCAL_HELP = 'Initializes Trestle workspace for local management of OSCAL models.'

FILE_ENCODING = 'utf8'

# Trestle documentation

WEBSITE_ROOT = 'https://oscal-compass.github.io/compliance-trestle'

BUG_REPORT = 'https://github.com/oscal-compass/compliance-trestle/issues/new/choose'

# Sample objects
SAMPLE_UUID_STR = 'A0000000-0000-4000-8000-000000000000'

WINDOWS_PLATFORM_STR = 'Windows'

# constants related to cache

TRESTLE_CACHE_DIR = TRESTLE_CONFIG_DIR + '/cache'

HOUR_SECONDS: int = 3600

DAY_SECONDS: int = 24 * HOUR_SECONDS

FILE_URI = 'file:///'

SFTP_URI = 'sftp://'

HTTPS_URI = 'https://'

# extracts letter, colon and single slash
WINDOWS_DRIVE_URI_REGEX = r'([A-Za-z]:[\\/]?)[^\\/]'

WINDOWS_DRIVE_LETTER_REGEX = r'[A-Za-z]:'

# extract two groups corresponding to prop name and value
# need to strip white space from end of value
PROPERTY_REGEX = r'(?:###\s+Property\s+)([^\s]*)\s*:\s+(.*)'

PART_REGEX = r'(?i)(?:##\s+part\s+)(.*)'

CONTROL_REGEX = r'(?i)(?:##\s+Control\s+)(.*)'

AFTER_HASHES_REGEX = r'(?:##*\s+)(.*)'

CACHE_ABS_DIR = '__abs__'

UNIX_CACHE_ROOT = '__root__'

TRESTLE_HREF_HEADING = 'trestle://'

TRESTLE_HREF_REGEX = '^trestle://[^/]'

MATCH_ALL_EXCEPT_LETTERS_UNDERSCORE_SPACE_REGEX = '[^a-zA-Z0-9-_ \n]'

# extracts foo and bar from ...[foo](bar)...
MARKDOWN_URL_REGEX = r'\[([^\]]+)\]\(([^)]+)\)'

# Governed header template version
TEMPLATE_VERSION_REGEX = r'[0-9]+.[0-9]+.[0-9]+'

OBJECTIVE_PART = 'objective'
ASSESMENT_OBJECTIVE_PART = 'assessment-objective'
TABLE_OF_PARAMS_PART = 'table_of_parameters'

# extracts standalone uuid's from anywhere in string
UUID_REGEX = r'(?:^|[0-9A-Za-f])([0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12})(?:$|[^0-9A-Za-z])'  # noqa FS003 E501

SSP_MAIN_COMP_NAME = 'This System'

THIS_SYSTEM_AS_KEY = 'this-system'

SSP_MD_HRULE_LINE = '______________________________________________________________________'

SSP_MD_IMPLEMENTATION_QUESTION = 'What is the solution and how is it implemented?'

SSP_MD_LEAVE_BLANK_TEXT = '<!-- Please leave this section blank and enter implementation details in the parts below. -->'  # noqa E501

SSP_ADD_IMPLEMENTATION_PREFIX = '<!-- Add control implementation description here for '

STATEMENT = 'statement'

CLOSE_COMMENT = ' -->'

SSP_ADD_IMPLEMENTATION_FOR_STATEMENT_TEXT = SSP_ADD_IMPLEMENTATION_PREFIX + STATEMENT + CLOSE_COMMENT

SSP_ADD_IMPLEMENTATION_FOR_ITEM_TEXT = SSP_ADD_IMPLEMENTATION_PREFIX + ITEM

SSP_ADD_IMPLEMENTATION_FOR_CONTROL_TEXT = SSP_ADD_IMPLEMENTATION_PREFIX + 'control'

SSP_ADD_THIS_SYSTEM_IMPLEMENTATION_FOR_CONTROL_TEXT = f'<!-- Add implementation prose for the main {SSP_MAIN_COMP_NAME} component for control'  # noqa E501

SSP_SYSTEM_CONTROL_IMPLEMENTATION_TEXT = 'This is the control implementation for the system.'

SSP_VALUES = 'ssp-values'

PROFILE_ADD_REQUIRED_SECTION_FOR_CONTROL_TEXT = '<!-- Add prose here for required Section'

PROFILE = 'profile'

LEVERAGED_SSP = 'leveraged-ssp'

TITLE = 'title'

NAME = 'name'

HREF = 'href'

NCNAME_REGEX = r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'  # noqa FS003 E501

NCNAME_UTF8_FIRST_CHAR_OPTIONS = string.ascii_letters + '_'

NCNAME_UTF8_OTHER_CHAR_OPTIONS = string.ascii_letters + string.digits + '_.-'

DISPLAY_VERBOSE_OUTPUT = 'Display verbose output'

HELP_YAML_PATH = 'Path to the optional yaml header file'
HELP_FO_OUTPUT = 'Overwrites the content of all markdowns in the output folder.'

HELP_OVERWRITE_HEADER_VALUES = (
    'Flag to overwrite values in a markdown control header.'
    + ' If a separate yaml header is passed in with -y, any items in the markdown header that are common with the'
    + ' provided header will be overwritten by the new values.  But new items passed in will always be added to the'
    + ' markdown header.'
)

HELP_REGENERATE = 'Flag to force generation of new uuids in the model'

HELP_VERSION = 'New version for the assembled model'

HELP_SECTIONS = 'Comma-separated list of sections as short_name_no_spaces:long name with spaces'

HELP_REQUIRED_SECTIONS = 'Short names of sections that must be in the assembled model, comma-separated'

HELP_ALLOWED_SECTIONS = 'Short names of sections that are allowed to be in the assembled model, comma-separated'

HELP_MARKDOWN_NAME = 'Name of the output generated profile markdown folder'

HELP_COMPDEFS = 'Comma-separated list of component-definitions for the ssp.'

FILTER_BY_PROFILE = 'filter-by-profile'

FILTER_BY_COMPONENTS = 'filter-by-components'

FILTER_EXCLUDE_COMPONENTS = 'filter-exclude-components'

GENERATE_RESOLVED_CATALOG = 'generate-resolved-catalog'

TRANSFORM_TYPES = [FILTER_BY_PROFILE, FILTER_BY_COMPONENTS, FILTER_EXCLUDE_COMPONENTS, GENERATE_RESOLVED_CATALOG]

TRESTLE_TAG = 'x-trestle-'

TRESTLE_PROPS_TAG = TRESTLE_TAG + 'props'

NAMESPACE_NIST = 'https://csrc.nist.gov/ns/oscal'

SSP_FEDRAMP_TAG = TRESTLE_TAG + 'fedramp-props'

TRESTLE_GLOBAL_TAG = TRESTLE_TAG + 'global'

NAMESPACE_FEDRAMP = 'https://fedramp.gov/ns/oscal'

LEV_AUTH_UUID = 'leveraged-authorization-uuid'

INHERITED_UUID = 'inherited-uuid'

IMPLEMENTATION_POINT = 'implementation-point'

IMPLEMENTATION_POINT_EXTERNAL = 'external'

STATUS_INHERITED = 'inherited'

STATUS_PARTIALLY_IMPLEMENTED = 'partially-implemented'

STATUS_PLANNED_COMPLETION_DATE = 'planned-completion-date'

STATUS_COMPLETION_DATE = 'completion-date'

CONTROL_ORIGINATION = 'control-origination'

IMPLEMENTATION_STATUS = 'implementation-status'

IMPLEMENTATION_STATUS_HEADER = 'Implementation Status'

IMPLEMENTATION_STATUS_REMARKS_HEADER = 'Implementation Status Remarks'

REMARKS = 'Remarks'

STATUS_REMARKS = 'status-remarks'

PROVIDED_STATEMENT_DESCRIPTION = 'Provided Statement Description'

RESPONSIBILITY_STATEMENT_DESCRIPTION = 'Responsibility Statement Description'

# Following 5 are allowed state tokens for
# SSP -> ControlImplementation -> ImplementedRequirements -> ByComponents -> common.ImplementationStatus -> State
# Also                         -> ImplementedRequirements -> Statements -> ByComponents ...
# But NIST says they may also be locally defined
STATUS_IMPLEMENTED = 'implemented'

STATUS_PARTIAL = 'partial'

STATUS_PARTIALLY_IMPLEMENTED = 'partially-implemented'

STATUS_PLANNED = 'planned'

STATUS_ALTERNATIVE = 'alternative'

STATUS_NOT_APPLICABLE = 'not-applicable'

# Following 4 needed by SSP -> SystemImplementation -> SystemComponent -> Status -> State1
# and by SSP -> SystemCharacteristics -> Status1 -> State
STATUS_OPERATIONAL = 'operational'

STATUS_UNDER_DEVELOPMENT = 'under-development'

STATUS_DISPOSITION = 'disposition'

STATUS_OTHER = 'other'

# Needed only by SystemCharacteristics
STATUS_UNDER_MAJOR_MODIFICATION = 'under-major-modification'

STATUS_ALL = [
    STATUS_IMPLEMENTED,
    STATUS_PARTIAL,
    STATUS_PARTIALLY_IMPLEMENTED,
    STATUS_PLANNED,
    STATUS_ALTERNATIVE,
    STATUS_NOT_APPLICABLE,
    STATUS_OPERATIONAL,
    STATUS_UNDER_DEVELOPMENT,
    STATUS_DISPOSITION,
    STATUS_OTHER,
    STATUS_UNDER_MAJOR_MODIFICATION
]

STATUS_PROMPT = f'<!-- For implementation status enter one of: {STATUS_IMPLEMENTED}, {STATUS_PARTIAL}, {STATUS_PLANNED}, {STATUS_ALTERNATIVE}, {STATUS_NOT_APPLICABLE} -->'  # noqa E501

RULES_WARNING = '<!-- Note that the list of rules under ### Rules: is read-only and changes will not be captured after assembly to JSON -->'  # noqa E501

SATISFIED_STATEMENT_COMMENT = (
    '<!-- Use this section to explain how'
    ' the inherited responsibility is being satisfied. -->'
)

THIS_SYSTEM_PROMPT = '### ' + SSP_MAIN_COMP_NAME

RESPONSIBLE_ROLE = 'responsible-role'

RESPONSIBLE_ROLES = 'responsible-roles'

HELP_SET_PARAMS = 'set parameters and properties based on the yaml header in control markdown'

SET_PARAMS_TAG = TRESTLE_TAG + 'set-params'

RULES_PARAMS_TAG = TRESTLE_TAG + 'rules-params'

COMP_DEF_RULES_PARAM_VALS_TAG = TRESTLE_TAG + 'comp-def-rules-param-vals'

TRESTLE_LEVERAGING_COMP_TAG = TRESTLE_TAG + 'leveraging-comp'

TRESTLE_STATEMENT_TAG = TRESTLE_TAG + 'statement'

PARAM_VALUES_TAG = TRESTLE_TAG + 'param-values'

COMP_DEF_RULES_TAG = TRESTLE_TAG + 'comp-def-rules'

PROFILE_VALUES = 'profile-values'

COMPONENT_VALUES = 'component-values'

VALUES = 'values'

GUIDELINES = 'guidelines'

PARAM_VALUE_ORIGIN = 'param-value-origin'

PROFILE_PARAM_VALUE_ORIGIN = 'profile-param-value-origin'

LABEL = 'label'

SECTIONS_TAG = TRESTLE_TAG + 'sections'

EDITABLE_CONTENT = 'Editable Content'

SORT_ID = 'sort-id'

TRESTLE_IMP_STATUS_TAG = 'trestle-imp-status'

TRESTLE_ADD_PROPS_TAG = TRESTLE_TAG + 'add-props'

TRESTLE_INHERITED_PROPS_TAG = TRESTLE_TAG + 'inherited-props'

CONTROL_OBJECTIVE_HEADER = '## Control Objective'

CONTROL_STATEMENT_HEADER = '## Control Statement'

CONTROL_HEADER = '## Control'

REPLACE_ME = 'REPLACE_ME'

PROVIDED_UUID = 'provided-uuid'

RESPONSIBILITY_UUID = 'responsibility-uuid'

YAML_PROPS_COMMENT = """  # Add or modify control properties here
  # Properties may be at the control or part level
  # Add control level properties like this:
  #   - name: ac1_new_prop
  #     value: new property value
  #
  # Add properties to a statement part like this, where "b." is the label of the target statement part
  #   - name: ac1_new_prop
  #     value: new property value
  #     smt-part: b.
  #
"""

YAML_SSP_VALUES_COMMENT = """  # You may set values for parameters in the assembled SSP by adding
  #
  # ssp-values:
  #   - value 1
  #   - value 2
  #
  # below a section of values:
  # The values list refers to the values in the resolved profile catalog, and the ssp-values represent new values
  # to be placed in SetParameters of the SSP.
  #
"""

YAML_PROFILE_VALUES_COMMENT = """    # This section contains the parameters that are part of this control.
  # Each parameter has properties. Only the profile-values and display-name properties are editable.
  # The other properties are informational.
  #
  # The values property for a parameter represents values inherited from the OSCAL catalog.
  # To override the catalog settings, use bullets under profile-values as shown below:
  #
  #   profile-values:
  #     - value 1
  #     - value 2
  #
  # If the "- <REPLACE_ME>" placeholder appears under profile-values, it is the same as if
  # the profile-values property were left empty.
  #
  # Some parameters may show an aggregates property which lists other parameters. This means
  # the parameter value is made up of the values from the other parameters. For parameters
  # that aggregate, profile-values is not applicable.
  #
  # Property param-value-origin is meant for putting the origin from where that parameter comes from.
  # In order to be changed in the current profile, profile-param-value-origin property will be displayed with
  # the placeholder "<REPLACE_ME>" for you to be replaced. If a parameter already has a param-value-origin
  # coming from an inherited profile, do no change this value, instead use profile-param-value-origin as follows:
  #
  #    param-value-origin: DO NOT REPLACE - this is the original value
  #    profile-param-value-origin: <REPLACE_ME> - replace the new value required HERE
  #
"""

YAML_RULE_PARAM_VALUES_SSP_COMMENT = """  # You may set new values for rule parameters by adding
  #
  # ssp-values:
  #   - value 1
  #   - value 2
  #
  # below a section of values:
  # The values list refers to the values as set by the components, and the ssp-values are the new values
  # to be placed in SetParameters of the SSP.
  #
"""

YAML_RULE_PARAM_VALUES_COMPONENT_COMMENT = """  # You may set new values for rule parameters by adding
  #
  # component-values:
  #   - value 1
  #   - value 2
  #
  # below a section of values:
  # The values list refers to the values as set by the components, and the component-values are the new values
  # to be placed in SetParameters of the component definition.
  #
"""

YAML_LEVERAGED_COMMENT = """  # Add or modify leveraged SSP Statements here.
"""

YAML_LEVERAGING_COMP_COMMENT = """  # Leveraged statements can be optionally associated with components in this system.
  # Associate leveraged statements to Components of this system here:
"""

DISPLAY_NAME = 'display-name'

TRESTLE_GENERIC_NS = 'https://oscal-compass.github.io/compliance-trestle/schemas/oscal'

RESOLUTION_SOURCE = 'resolution-source'

# call it tracker to distinguish from the externally visible TRESTLE_INHERITED_PROPS_TAG
TRESTLE_INHERITED_PROPS_TRACKER = 'trestle_inherited_props_tracker'

RULE_ID = 'Rule_Id'

HEADER_RULE_ID = 'rule-id'

RULE_DESCRIPTION = 'Rule_Description'

PARAMETER_ID = 'Parameter_Id'

PARAMETER_DESCRIPTION = 'Parameter_Description'

PARAMETER_VALUE_ALTERNATIVES = 'Parameter_Value_Alternatives'

ONE = 'one'

ONE_OR_MORE_HYPHENED = 'one-or-more'

ONE_OR_MORE_SPACED = 'one or more'

VALUE_ASSIGNED_PREFIX = 'value-assigned-prefix'

VALUE_NOT_ASSIGNED_PREFIX = 'value-not-assigned-prefix'

CONTROL_IMPLEMENTATION = 'control-implementation'

AGGREGATES = 'aggregates'

ALT_IDENTIFIER = 'alt-identifier'

IMPLEMENTED_REQUIREMENT = 'implemented-requirement'

# Following 5 are allowed control origination values for
# SSP -> ControlImplementation -> ImplementedRequirements -> prop[@name='control-origination']/@value
ORIGINATION_ORGANIZATION = 'organization'

ORIGINATION_SYSTEM_SPECIFIC = 'system-specific'

ORIGINATION_CUSTOMER_CONFIGURED = 'customer-configured'

ORIGINATION_CUSTOMER_PROVIDED = 'customer-provided'

ORIGINATION_INHERITED = 'inherited'

# Constant relation to the inheritance view Markdown

INHERITANCE_VIEW_DIR = 'inheritance'

HELP_LEVERAGED = 'Name of the SSP to be leveraged.'

SATISFIED_STATEMENT_DESCRIPTION = 'Satisfied Statement Description'

ADDED_BY_CONTROL_OWNER = 'Added by control owner'

REPLACE_ME_PLACEHOLDER = '<REPLACE_ME>'



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Jinja command."""
import os
import pathlib
import shutil

from _pytest.monkeypatch import MonkeyPatch

from tests.test_utils import execute_command_and_assert, setup_for_ssp

from trestle.core.commands.author.jinja import _number_captions
from trestle.core.commands.author.ssp import SSPGenerate
from trestle.core.markdown.docs_markdown_node import DocsMarkdownNode


def setup_ssp(
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch,
    profile_name: str = 'comp_prof'
):
    """Prepare repository for docs generation."""
    args, _ = setup_for_ssp(tmp_trestle_dir, profile_name, 'my_ssp')
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0

    command_ssp_gen = f'trestle author ssp-assemble -m my_ssp -o ssp_json -cd {args.compdefs}'
    execute_command_and_assert(command_ssp_gen, 0, monkeypatch)

    for file_name in os.listdir(testdata_dir / 'jinja'):
        full_file_name = os.path.join(testdata_dir / 'jinja', file_name)
        if os.path.isfile(full_file_name):
            shutil.copy(full_file_name, tmp_trestle_dir)


def test_jinja_ssp_output(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test Jinja SSP output."""
    input_template = 'ssp_template.md.jinja'
    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)
    command_import = f'trestle author jinja -i {input_template} -o output_file.md -ssp ssp_json -p comp_prof'
    execute_command_and_assert(command_import, 0, monkeypatch)

    with open('output_file.md') as test_output:
        output = test_output.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(output.split('\n'))
        assert tree
        node1 = tree.get_node_for_key('# A')
        node2 = tree.get_node_for_key('# C')
        assert node1.subnodes[0].key == node2.subnodes[0].key
        assert len(node1.subnodes) == len(node2.subnodes)


def test_jinja_lookup_table(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test Jinja lookup table substitions."""
    input_template = 'use_lookup_table.md.jinja'
    luk_table = 'lookup_table.yaml'
    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)
    command_import = f'trestle author jinja -i {input_template} -o output_file.md ' \
                     f'-lut {luk_table} -ssp ssp_json -p comp_prof -elp lut.prefix'  # noqa: N400
    execute_command_and_assert(command_import, 0, monkeypatch)

    with open('output_file.md') as test_output:
        output = test_output.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(output.split('\n'))
        assert tree
        node1 = tree.get_node_for_key('### Substitutions')
        node2 = tree.get_node_for_key('# B')
        assert node1.content.text[1] == 'This word: compliance-trestle, was substituted.'
        assert node2.content.text


def test_number_captions(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test numbering of captions in markdown."""
    with open(testdata_dir / 'jinja_cmd/number_captions_data.md', 'r') as fp:
        test_data = fp.read()

    with open(testdata_dir / 'jinja_cmd/number_captions_expected_output.md', 'r') as fp:
        expected_output = fp.read().splitlines()

    for idx, row in enumerate(_number_captions(test_data).splitlines()):
        assert row == expected_output[idx]


def test_params_formatting(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that parameters are substituted with the given formatting."""
    input_template = 'ssp_template.md.jinja'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_md_gen = f'trestle author jinja -bf *.* -vap "IBM Assignment:" -vnap "Assignment:" -i {input_template} -o output.md -ssp ssp_json -p comp_prof'  # noqa E501
    execute_command_and_assert(command_md_gen, 0, monkeypatch)

    with open('output.md') as test_output:
        output = test_output.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(output.split('\n'))
        assert tree
        parent = tree.get_node_for_key('### AC-1 - Policy and Procedures')
        child1 = parent.get_node_for_key('#### AC-1 Summary information')
        assert child1
        child2 = parent.get_node_for_key('#### Control Statement')
        assert '*Assignment: organization-defined official*' in child2.content.raw_text

    command_md_gen = f'trestle author jinja -bf Prefix:[.] -vap "IBM Assignment:" -vnap "Assignment:" -i {input_template} -o output.md -ssp ssp_json -p comp_prof'  # noqa E501
    execute_command_and_assert(command_md_gen, 0, monkeypatch)

    with open('output.md') as test_output:
        output = test_output.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(output.split('\n'))
        assert tree
        parent = tree.get_node_for_key('### AC-1 - Policy and Procedures')
        child1 = parent.get_node_for_key('#### AC-1 Summary information')
        child2 = parent.get_node_for_key('#### Control Statement')
        assert child1
        assert 'Prefix:[Assignment: organization-defined official]' in child2.content.raw_text


def test_jinja_profile_docs(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test Jinja Profile to multiple md files output."""
    input_template = 'profile_to_docs.md.jinja'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_import = f'trestle author jinja -i {input_template} -o controls -p comp_prof --docs-profile'
    execute_command_and_assert(command_import, 0, monkeypatch)

    for md_control in (tmp_trestle_dir / 'controls' / 'ac').iterdir():
        with open(md_control) as md_file:
            contents = md_file.read()
            tree = DocsMarkdownNode.build_tree_from_markdown(contents.split('\n'))
            assert tree
            node1 = tree.get_node_for_key('# Control Page')
            assert node1
            node2 = tree.get_node_for_key('## Statement Header')
            assert node2
            if md_control.name == 'ac-1.md':
                node3 = tree.get_node_for_key('## Implementation Guidance')
                assert node3

                node4 = tree.get_node_for_key('# AC-1 - Policy and Procedures')
                assert node4.get_node_for_key('## Implementation Guidance')


def test_jinja_profile_docs_no_part_prose(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test Jinja Profile to multiple md files output. Test for if part to add does not have prose at all."""
    input_template = 'profile_to_docs_no_part_prose.md.jinja'
    profile_name = 'comp_prof_part_none'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch, profile_name)

    command_import = f'trestle author jinja -i {input_template} -o controls -p {profile_name} --docs-profile'
    execute_command_and_assert(command_import, 0, monkeypatch)


def test_jinja_profile_docs_with_group_title(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test Jinja Profile to multiple md files output with group title."""
    input_template = 'profile_to_docs_with_group_title.md.jinja'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_import = f'trestle author jinja -i {input_template} -o controls -p comp_prof --docs-profile'
    execute_command_and_assert(command_import, 0, monkeypatch)

    md_control = tmp_trestle_dir / 'controls' / 'ac' / 'ac-1.md'
    with open(md_control) as md_file:
        contents = md_file.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(contents.split('\n'))
        assert tree
        node1 = tree.get_node_for_key('# Control Page')
        assert node1
        node2 = tree.get_node_for_key('# AC-1 - \\[Access Control\\] Policy and Procedures')
        assert node2
        assert '{: #ac-1}' in node2.content.raw_text  # noqa: FS003 - not f string but tag
        assert node2.content.text[1] == ''  # assert new line after tag
        node3 = tree.get_node_for_key('## Table of Control Parameters')
        assert node3
        assert '{: #table-of-control-parameters}' in node3.content.raw_text  # noqa: FS003 - not f string but tag
        assert '{: #"Parameters for AC-1" caption-side="top"}' in node3.content.raw_text  # noqa: FS003 - not f string
        assert 'AC-1 (a) (1)' in node3.content.tables[2]
        assert 'ac-1_prm_3' in node3.content.tables[4]


def test_jinja_profile_docs_with_selected_sections(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test Jinja Profile to multiple md files output with group title."""
    input_template = 'profile_to_docs_only_some_sections.md.jinja'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_import = f'trestle author jinja -i {input_template} -o controls -p comp_prof --docs-profile'
    execute_command_and_assert(command_import, 0, monkeypatch)

    md_control = tmp_trestle_dir / 'controls' / 'ac' / 'ac-1.md'
    with open(md_control) as md_file:
        contents = md_file.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(contents.split('\n'))
        assert tree
        node1 = tree.get_node_for_key('# Control Page')
        assert node1
        node2 = tree.get_node_for_key('## Control Statement Header')
        assert node2
        assert '{: #control-statement-header}' in node2.content.raw_text  # noqa: FS003 - not f string but tag
        assert len(tree.content.subnodes_keys) == 3


def test_jinja_profile_docs_with_selected_sections_and_multiple_parts(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test Jinja Profile to multiple md files output with selected sections and multiple subparts."""
    input_template = 'profile_to_docs_with_subparts.md.jinja'
    profile_path = testdata_dir / 'json/profile_with_alter_subparts.json'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_import = f'trestle import -f {profile_path} -o myprofile'
    execute_command_and_assert(command_import, 0, monkeypatch)

    command_jinja = f'trestle author jinja -i {input_template} -o controls -p myprofile --docs-profile'
    execute_command_and_assert(command_jinja, 0, monkeypatch)

    md_control = tmp_trestle_dir / 'controls' / 'ac' / 'ac-1.md'
    with open(md_control) as md_file:
        contents = md_file.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(contents.split('\n'))
        assert tree
        node1 = tree.get_node_for_key('## The above the line guidance')
        assert node1
        node2 = tree.get_node_for_key('### Add to part a')
        assert node2
        node3 = tree.get_node_for_key('#### Evidence Guidance')
        assert node3
        tag = '{: #the-above-the-line-guidance-add-to-part-a-evidence-guidance}'  # noqa: FS003
        assert tag in node2.content.raw_text
        assert len(tree.content.subnodes_keys) == 7


def test_jinja_profile_docs_fails(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test jinja docs generate fails."""
    input_template = 'profile_to_docs.md.jinja'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_jinja = f'trestle author jinja -i {input_template} -o controls --docs-profile'
    execute_command_and_assert(command_jinja, 2, monkeypatch)

    command_jinja = f'trestle author jinja -i {input_template} -o controls -ssp ssp_json -p main_profile --docs-profile'
    execute_command_and_assert(command_jinja, 2, monkeypatch)

    input_template = 'profile_to_docs_invalid.md.jinja'
    command_jinja = f'trestle author jinja -i {input_template} -o controls -p main_profile --docs-profile'
    execute_command_and_assert(command_jinja, 1, monkeypatch)


def test_jinja_with_template_only(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Test jinja output with only input template.

    Oscal profile, docs profile and ssp have not been provided as input.
    """
    input_template = 'ssp_template_no_input_profile_ssp.md.jinja'

    setup_ssp(testdata_dir, tmp_trestle_dir, monkeypatch)

    command_jinja = f'trestle author jinja -i {input_template} -o output.md -v'
    execute_command_and_assert(command_jinja, 0, monkeypatch)

    with open('output.md', 'r') as md_file:
        content = md_file.read()
        tree = DocsMarkdownNode.build_tree_from_markdown(content.split('\n'))
        assert tree
        node1 = tree.get_node_for_key('# A')
        node2 = tree.get_node_for_key('# C')
        assert node1.subnodes[0].key == node2.subnodes[0].key



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Init Command."""
import argparse
import logging
import pathlib
import shutil
from shutil import copyfile

from pkg_resources import resource_filename

import trestle.common.const as const
import trestle.common.log as log
from trestle.common import file_utils
from trestle.common.err import TrestleError, TrestleRootError, handle_generic_command_exception
from trestle.core.commands.command_docs import CommandBase
from trestle.core.commands.common.return_codes import CmdReturnCodes

logger = logging.getLogger(__name__)


class InitCmd(CommandBase):
    """Initialize a trestle working directory."""

    name = 'init'

    def _init_arguments(self) -> None:

        self.add_argument(const.INIT_FULL_SHORT, const.INIT_FULL_LONG, help=const.INIT_FULL_HELP, action='store_true')
        self.add_argument(
            const.INIT_LOCAL_SHORT, const.INIT_LOCAL_LONG, help=const.INIT_LOCAL_HELP, action='store_true'
        )
        self.add_argument(
            const.INIT_GOVDOCS_SHORT, const.INIT_GOVDOCS_LONG, help=const.INIT_GOVDOCS_HELP, action='store_true'
        )

    def _run(self, args: argparse.Namespace) -> int:
        """Create a trestle workspace in the current directory."""
        try:
            log.set_log_level_from_args(args)
            dir_path: pathlib.Path = args.trestle_root
            if not dir_path.exists() or not dir_path.is_dir():
                raise TrestleRootError(
                    f'Initialization failed. Given directory {dir_path} does not exist or is not a directory.'
                )

            if not (args.full or args.local or args.govdocs):
                # Running in full mode by default
                args.full = True

            init_dist_folder = True if args.full else False
            init_oscal_folders = True if args.local or args.full else False
            copy_config_file = True if args.full or args.local else False

            self._create_directories(dir_path, init_oscal_folders, init_dist_folder)

            if copy_config_file:
                self._copy_config_file(dir_path)

            logger.info(f'Initialized trestle project successfully in {dir_path}')

            return CmdReturnCodes.SUCCESS.value

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Failed to initialize Trestle working directory.')

    def _create_directories(self, root: pathlib.Path, create_model_folders: bool, create_dist_folder: bool) -> None:
        """Create the directory tree if it does not exist."""
        # Prepare directory list to be created
        try:
            directory_list = [root / pathlib.Path(const.TRESTLE_CONFIG_DIR)]

            if create_model_folders:
                for model_dir in const.MODEL_DIR_LIST:
                    directory_list.append(root / pathlib.Path(model_dir))
                    if create_dist_folder:
                        directory_list.append(root / pathlib.Path(const.TRESTLE_DIST_DIR) / model_dir)

            # Create directories
            for directory in directory_list:
                directory.mkdir(parents=True, exist_ok=True)
                file_path = pathlib.Path(directory) / const.TRESTLE_KEEP_FILE
                file_utils.make_hidden_file(file_path)
        except OSError as e:
            raise TrestleError(f'Error while creating directories: {e}')
        except Exception as e:
            raise TrestleError(f'Unexpected error while creating directories: {e}')

    def _copy_config_file(self, root: pathlib.Path) -> None:
        """Copy the initial config.ini file to .trestle directory."""
        try:
            source_path = pathlib.Path(resource_filename('trestle.resources', const.TRESTLE_CONFIG_FILE)).resolve()
            destination_path = (root / pathlib.Path(const.TRESTLE_CONFIG_DIR) / const.TRESTLE_CONFIG_FILE).resolve()
            copyfile(source_path, destination_path)

        except (shutil.SameFileError, OSError) as e:
            raise TrestleError(f'Error while copying config file: {e}')
        except Exception as e:
            raise TrestleError(f'Unexpected error while copying config file: {e}')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Describe Command."""

import argparse
import logging
import pathlib
from typing import List, Optional, Union

import trestle.common.log as log
from trestle.common.err import TrestleError, TrestleIncorrectArgsError, handle_generic_command_exception
from trestle.common.model_utils import ModelUtils
from trestle.core.base_model import OscalBaseModel
from trestle.core.commands.command_docs import CommandPlusDocs
from trestle.core.commands.common import cmd_utils as utils
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.models.elements import Element

logger = logging.getLogger(__name__)


class DescribeCmd(CommandPlusDocs):
    """Describe contents of a model file including optional element path."""

    # The only output is via log lines.  No other results or side-effects.

    name = 'describe'

    def _init_arguments(self) -> None:
        logger.debug('Init arguments')
        self.add_argument('-f', '--file', help='OSCAL file to import.', type=str, required=True)

        self.add_argument(
            '-e', '--element', help='Optional name of element in file to describe.', type=str, required=False
        )

    def _run(self, args: argparse.Namespace) -> int:
        try:
            logger.debug('Entering trestle describe.')

            log.set_log_level_from_args(args)

            if args.file:
                model_file = pathlib.Path(args.file)

                element = '' if not args.element else args.element.strip("'")
                results = self.describe(model_file.resolve(), element, args.trestle_root)

                return CmdReturnCodes.SUCCESS.value if len(results) > 0 else CmdReturnCodes.COMMAND_ERROR.value
            else:
                raise TrestleIncorrectArgsError('No file specified for command describe.')

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while describing contents of a model')

    @classmethod
    def _clean_type_string(cls, text: str) -> str:
        text = text.replace("<class '", '').replace("'>", '')
        text = text.replace('trestle.oscal.', '')
        text = text.replace('pydantic.main.', 'stripped.')
        return text

    @classmethod
    def _description_text(cls, sub_model: Optional[Union[OscalBaseModel, List[OscalBaseModel]]]) -> str:
        clip_string = 100
        if sub_model is None:
            return 'None'
        if type(sub_model) is list:
            n_items = len(sub_model)
            type_text = 'Unknown' if not n_items else f'{cls._clean_type_string(str(type(sub_model[0])))}'
            text = f'list of {n_items} items of type {type_text}'
            return text
        if type(sub_model) is str:
            return sub_model if len(sub_model
                                    ) < clip_string else sub_model[:clip_string] + '[truncated]'  # type: ignore
        if hasattr(sub_model, 'type_'):
            return cls._clean_type_string(str(sub_model.type_))
        return cls._clean_type_string(str(type(sub_model)))

    @classmethod
    def describe(cls, file_path: pathlib.Path, element_path_str: str, trestle_root: pathlib.Path) -> List[str]:
        """Describe the contents of the file.

        Args:
            file_path: pathlib.Path Path for model file to describe.
            element_path_str: Element path of element in model to describe.  Can be ''.

        Returns:
            The list of lines of text in the description, or an empty list on failure
        """
        # figure out the model type so we can read it
        try:
            model_type, _ = ModelUtils.get_stripped_model_type(file_path, trestle_root)
            model: OscalBaseModel = model_type.oscal_read(file_path)
        except TrestleError as e:
            logger.warning(f'Error loading model {file_path} to describe: {e}')
            return []

        sub_model = model

        # if an element path was provided, follow the path chain to the desired sub_model
        if element_path_str:
            if '*' in element_path_str or ',' in element_path_str:
                logger.warning('Wildcards and commas are not allowed in the element path for describe.')
                return []

            if '.' not in element_path_str:
                logger.warning('The element path for describe must either be omitted or contain at least 2 parts.')
                return []

            element_paths = utils.parse_element_arg(model, element_path_str)

            sub_model_element = Element(model)

            for element_path in element_paths:
                sub_model = sub_model_element.get_at(element_path, False)
                sub_model_element = Element(sub_model)

        # now that we have the desired sub_model we can describe it

        text_out: List[str] = []

        # create top level text depending on whether an element path was used
        element_text = '' if not element_path_str else f' at element path {element_path_str}'

        if type(sub_model) is list:
            text = f'Model file {file_path}{element_text} is a {cls._description_text(sub_model)}'
            text_out.append(text)
            logger.info(text)
        else:
            text = f'Model file {file_path}{element_text} is of type '
            text += f'{cls._clean_type_string(str(type(sub_model)))} and contains:'
            text_out.append(text)
            logger.info(text)
            for key in sub_model.__fields__.keys():
                value = getattr(sub_model, key, None)
                text = f'    {key}: {cls._description_text(value)}'
                text_out.append(text)
                logger.info(text)

        return text_out



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL transformation tasks."""

# mypy: ignore-errors  # noqa E800
import configparser
import datetime
import logging
import pathlib
import traceback
import uuid
from collections import OrderedDict
from typing import Iterator, List, Optional

from openpyxl import load_workbook

from trestle.oscal import OSCAL_VERSION
from trestle.oscal.catalog import Catalog
from trestle.oscal.catalog import Control
from trestle.oscal.catalog import Group
from trestle.oscal.common import BackMatter
from trestle.oscal.common import Link
from trestle.oscal.common import Metadata
from trestle.oscal.common import Part
from trestle.oscal.common import Property
from trestle.oscal.common import Resource
from trestle.tasks.base_task import TaskBase
from trestle.tasks.base_task import TaskOutcome

logger = logging.getLogger(__name__)

timestamp = datetime.datetime.utcnow().replace(microsecond=0).replace(tzinfo=datetime.timezone.utc).isoformat()


class XlsxHelper:
    """Xlsx Helper common functions and assistance navigating spread sheet."""

    def __init__(self, file: str) -> None:
        """Initialize."""
        self._spread_sheet = file
        self._wb = load_workbook(self._spread_sheet)
        sheet_candidates = ['Combined Profiles', 'Combined']
        self._sheet_name = None
        for sheet_candidate in sheet_candidates:
            if sheet_candidate in self._wb.sheetnames:
                self._sheet_name = sheet_candidate
                break
        if not self._sheet_name:
            raise RuntimeError(f'{file} missing one of {sheet_candidates} sheet')
        self._work_sheet = self._wb[self._sheet_name]
        self._mapper()
        self._key_to_col_map = {'statement': 'description'}

    def is_ocp(self) -> bool:
        """Check if sheet is for OCP."""
        return self._sheet_name == 'Combined Profiles'

    def _normalize(self, name: str) -> str:
        """Normalize."""
        return name.lower()

    def _translate(self, name: str) -> str:
        """Translate name key to column name."""
        return self._key_to_col_map.get(name, name)

    def _mapper(self) -> None:
        """Map columns heading names to column numbers."""
        self._col_name_to_number = {}
        cols = self._work_sheet.max_column
        row = 1
        for col in range(row, cols):
            cell = self._work_sheet.cell(row, col)
            if cell.value:
                name = self._normalize(cell.value)
                self._col_name_to_number[name] = col

    def row_generator(self) -> Iterator[int]:
        """Generate rows until max reached."""
        row = 2
        while row <= self._work_sheet.max_row:
            yield row
            row += 1

    def get(self, row: int, name: str) -> str:
        """Get cell value for given row and column name."""
        nname = self._normalize(name)
        cname = self._translate(nname)
        col = self._col_name_to_number[cname]
        cell = self._work_sheet.cell(row, col)
        return cell.value


class CatalogHelper:
    """OSCAL Catalog Helper."""

    def __init__(self, title: str, version: str) -> None:
        """Initialize."""
        # metadata
        self._metadata = Metadata(title=title, last_modified=timestamp, oscal_version=OSCAL_VERSION, version=version)
        self._root_group = OrderedDict()
        self._root_resources = OrderedDict()
        self._all_groups = OrderedDict()
        self._all_controls = OrderedDict()

    def add_group(self, section: str, title: str, props: List[Property], parts: List[Part]) -> None:
        """Add group."""
        numdots = section.count('.')
        if numdots == 0:
            group = Group(title=f'{title}', id=f'CIS-{section}')
            if props:
                group.props = props
            if parts:
                group.parts = parts
            self._root_group[section] = group
            self._all_groups[section] = group
        else:
            key = '.'.join(section.split('.')[:-1])
            parent = self._all_groups[key]
            if parent.groups is None:
                parent.groups = []
            group = Group(title=f'{title}', id=f'CIS-{section}')
            if props:
                group.props = props
            if parts:
                group.parts = parts
            parent.groups.append(group)
            self._all_groups[section] = group

    def _add_prop(self, control: Control, prop: Property) -> None:
        """Add property to control."""
        control_props = control.props
        control.props = []
        last = 0
        for i, control_prop in enumerate(control_props):
            if control_prop.name == prop.name:
                last = i
        for i, control_prop in enumerate(control_props):
            control.props.append(control_prop)
            if i == last:
                control.props.append(prop)

    def add_control(
        self,
        section: str,
        recommendation: str,
        title: str,
        props: List[Property],
        parts: List[Part],
        links: List[Link]
    ) -> None:
        """Add control."""
        group = self._all_groups[section]
        if group.controls is None:
            group.controls = []
        id_ = f'CIS-{recommendation}'
        if id_ in self._all_controls:
            control = self._all_controls[id_]
            for prop in props:
                if prop.name == 'profile':
                    self._add_prop(control, prop)
        else:
            title = f'{title}'
            control = Control(id=id_, title=title)
            self._all_controls[id_] = control
            if props:
                control.props = props
            if parts:
                control.parts = parts
            if links:
                control.links = links
            group.controls.append(control)

    def add_link(
        self,
        recommendation: str,
        reference: str,
        links: List[Link],
    ) -> None:
        """Add link."""
        id_ = f'CIS-{recommendation}'
        if id_ not in self._root_resources:
            res_id = str(uuid.uuid4())
            link = Link(href=f'#{res_id}', rel='reference')
            links.append(link)
            resource = Resource(
                uuid=res_id,
                description=reference,
            )
            self._root_resources[id_] = resource

    def get_catalog(self) -> Catalog:
        """Get catalog."""
        back_matter = BackMatter(resources=list(self._root_resources.values()))
        catalog = Catalog(
            uuid=str(uuid.uuid4()),
            metadata=self._metadata,
            groups=list(self._root_group.values()),
            back_matter=back_matter
        )
        return catalog


class CisXlsxToOscalCatalog(TaskBase):
    """
    Task to transform CIS .xlsx to OSCAL catalog.

    Attributes:
        name: Name of the task.
    """

    name = 'cis-xlsx-to-oscal-catalog'
    ns = 'https://oscal-compass.github.io/compliance-trestle/schemas/oscal/catalog/cis'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task.

        Args:
            config_object: Config section associated with the task.
        """
        super().__init__(config_object)

    def print_info(self) -> None:
        """Print the help string."""
        logger.info(f'Help information for {self.name} task.')
        logger.info('')
        logger.info('Purpose: Create catalog from standard (e.g. CIS benchmark).')
        logger.info('')
        logger.info('Configuration flags sit under [task.cis-xlsx-to-oscal-catalog]:')
        text1 = '  input-file             = '
        text2 = '(required) path to read the compliance-as-code .xlsx spread sheet file.'
        logger.info(text1 + text2)
        text1 = '  output-dir             = '
        text2 = '(required) location to write the generated catalog.json file.'
        logger.info(text1 + text2)
        text1 = '  title                  = '
        text2 = '(required) title of the CIS catalog.'
        logger.info(text1 + text2)
        text1 = '  version                = '
        text2 = '(required) version of the CIS catalog.'
        logger.info(text1 + text2)
        text1 = '  output-overwrite       = '
        text2 = '(optional) true [default] or false; replace existing output when true.'
        logger.info(text1 + text2)

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        return TaskOutcome('simulated-success')

    def execute(self) -> TaskOutcome:
        """Provide an actual outcome."""
        try:
            return self._execute()
        except Exception:
            logger.info(traceback.format_exc())
            return TaskOutcome('failure')

    def _get_normalized_name(self, key: str) -> str:
        """Get normalized name."""
        name = key
        name = name.replace(' ', '_')
        name = name.replace('&', 'a')
        name = name.replace('(', '')
        name = name.replace(')', '')
        return name

    def _create_property(self, name: str, value: str) -> Property:
        return Property(name=name, value=value, ns=CisXlsxToOscalCatalog.ns)

    def _add_property(self, xlsx_helper: XlsxHelper, props: List[Property], row: int, key: str) -> None:
        """Add property."""
        name = self._get_normalized_name(key)
        value = xlsx_helper.get(row, key)
        if value:
            props.append(self._create_property(name, value))

    def _add_property_boolean(self, xlsx_helper: XlsxHelper, props: List[Property], row: int, key: str) -> None:
        """Add property."""
        name = self._get_normalized_name(key)
        value = xlsx_helper.get(row, key)
        if value:
            props.append(self._create_property(name, 'True'))
        else:
            props.append(self._create_property(name, 'False'))

    def _add_part(self, xlsx_helper: XlsxHelper, parts: List[Part], id_: str, row: int, key: str) -> None:
        """Add part."""
        value = xlsx_helper.get(row, key)
        if value:
            name = self._get_normalized_name(key)
            parts.append(Part(id=id_, name=name, prose=value))

    def _add_links(
        self, xlsx_helper: XlsxHelper, catalog_helper: CatalogHelper, links: List[Link], row: int, key: str
    ) -> None:
        """Add links."""
        value = xlsx_helper.get(row, key)
        if value:
            recommendation = xlsx_helper.get(row, 'recommendation #')
            catalog_helper.add_link(recommendation, value, links)

    def _execute(self) -> TaskOutcome:
        """Wrap the execute for exception handling."""
        if not self._config:
            logger.warning('config missing')
            return TaskOutcome('failure')
        try:
            ifile = self._config['input-file']
            odir = self._config['output-dir']
            title = self._config['title']
            version = self._config['version']
        except KeyError as e:
            logger.info(f'key {e.args[0]} missing')
            return TaskOutcome('failure')
        # verbosity
        _quiet = self._config.get('quiet', False)
        _verbose = not _quiet
        # output
        overwrite = self._config.getboolean('output-overwrite', True)
        opth = pathlib.Path(odir)
        # insure output dir exists
        opth.mkdir(exist_ok=True, parents=True)
        # calculate output file name & check writability
        oname = 'catalog.json'
        ofile = opth / oname
        if not overwrite and pathlib.Path(ofile).exists():
            logger.warning(f'output: {ofile} already exists')
            return TaskOutcome('failure')
        xlsx_helper = XlsxHelper(ifile)
        catalog_helper = CatalogHelper(title, version)
        if xlsx_helper.is_ocp():
            self._process_ocp(xlsx_helper, catalog_helper)
        else:
            self._process_rhel(xlsx_helper, catalog_helper)
        catalog = catalog_helper.get_catalog()
        # write OSCAL ComponentDefinition to file
        if _verbose:
            logger.info(f'output: {ofile}')
        catalog.oscal_write(pathlib.Path(ofile))
        return TaskOutcome('success')

    def _process_ocp(self, xlsx_helper: XlsxHelper, catalog_helper: CatalogHelper) -> None:
        """Process OCP."""
        # transform each row into OSCAL equivalent
        for row in xlsx_helper.row_generator():
            section = xlsx_helper.get(row, 'section #')
            recommendation = xlsx_helper.get(row, 'recommendation #')
            title = xlsx_helper.get(row, 'title')
            # init
            props = []
            parts = []
            links = []
            # props
            self._add_property(xlsx_helper, props, row, 'profile')
            self._add_property(xlsx_helper, props, row, 'status')
            self._add_property(xlsx_helper, props, row, 'assessment status')
            if recommendation is None:
                frag = section
            else:
                frag = recommendation
            # parts
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_smt', row, 'statement')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_rat', row, 'rationale statement')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_imp', row, 'impact statement')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_rem', row, 'remediation procedure')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_aud', row, 'audit procedure')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_inf', row, 'additional information')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_ctl', row, 'CIS Controls')
            # group or control
            if recommendation is None:
                catalog_helper.add_group(section, title, props, parts)
            else:
                self._add_property_boolean(xlsx_helper, props, row, 'v7 IG1')
                self._add_property_boolean(xlsx_helper, props, row, 'v7 IG2')
                self._add_property_boolean(xlsx_helper, props, row, 'v7 IG3')
                self._add_property_boolean(xlsx_helper, props, row, 'v8 IG1')
                self._add_property_boolean(xlsx_helper, props, row, 'v8 IG2')
                self._add_property_boolean(xlsx_helper, props, row, 'v8 IG3')
                self._add_property(xlsx_helper, props, row, 'MITRE ATT&CK Mappings')
                self._add_links(xlsx_helper, catalog_helper, links, row, 'references')
                catalog_helper.add_control(section, recommendation, title, props, parts, links)

    def _process_rhel(self, xlsx_helper: XlsxHelper, catalog_helper: CatalogHelper) -> None:
        """Process RHEL."""
        # transform each row into OSCAL equivalent
        for row in xlsx_helper.row_generator():
            section = xlsx_helper.get(row, 'section #')
            recommendation = xlsx_helper.get(row, 'recommendation #')
            title = xlsx_helper.get(row, 'title')
            # init
            props = []
            parts = []
            links = []
            # props
            self._add_property(xlsx_helper, props, row, 'profile')
            self._add_property(xlsx_helper, props, row, 'assessment status')
            if recommendation is None:
                frag = section
            else:
                frag = recommendation
            # parts
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_smt', row, 'statement')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_rat', row, 'rational statement')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_imp', row, 'impact statement')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_rem', row, 'remediation procedure')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_aud', row, 'audit procedure')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_inf', row, 'additional information')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_ctl', row, 'CIS Controls')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_1v8', row, 'CIS Safeguards 1 (v8)')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_2v8', row, 'CIS Safeguards 2 (v8)')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_3v8', row, 'CIS Safeguards 3 (v8)')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_1v7', row, 'CIS Safeguards 1 (v7)')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_2v7', row, 'CIS Safeguards 2 (v7)')
            self._add_part(xlsx_helper, parts, f'CIS-{frag}_3v7', row, 'CIS Safeguards 3 (v7)')
            # group or control
            if recommendation is None:
                catalog_helper.add_group(section, title, props, parts)
            else:
                self._add_property_boolean(xlsx_helper, props, row, 'v7 IG1')
                self._add_property_boolean(xlsx_helper, props, row, 'v7 IG2')
                self._add_property_boolean(xlsx_helper, props, row, 'v7 IG3')
                self._add_property_boolean(xlsx_helper, props, row, 'v8 IG1')
                self._add_property_boolean(xlsx_helper, props, row, 'v8 IG2')
                self._add_property_boolean(xlsx_helper, props, row, 'v8 IG3')
                self._add_links(xlsx_helper, catalog_helper, links, row, 'references')
                catalog_helper.add_control(section, recommendation, title, props, parts, links)



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for list_utils module."""

import pytest

from trestle.common import list_utils
from trestle.common.err import TrestleError


def test_is_ordered_sublist() -> None:
    """Test is_ordered_sublist method."""
    assert list_utils.is_ordered_sublist(['a', 'b', 'c'], ['x', 'a', 'b', 'c', 'y', 'z']) is True
    assert list_utils.is_ordered_sublist(['a', 'b', 'c'], ['x', 'a', 'c', 'b', 'y', 'z']) is False
    assert list_utils.is_ordered_sublist(['a', 'b', 'c'], ['x', 'a', 'b', 'y', 'c', 'z']) is False


def test_join_key_to_list_dicts() -> None:
    """Test join_key_to_list_dicts method."""
    d1 = {'a': [1, 2], 'b': [3, 4], 'c': [5], 'd': [8, 9]}
    d2 = {'a': [2, 1], 'b': [3], 'c': [6, 7], 'e': [0]}
    d3 = {'a': [1, 2, 2, 1], 'b': [3, 4, 3], 'c': [5, 6, 7], 'd': [8, 9], 'e': [0]}
    d4 = list_utils.join_key_to_list_dicts(d1, d2)
    assert d4 == d3
    assert d4 != d1
    d1 = {1: [1, 2], 2: [3, 4], '3': [5], 'd': [8, 9]}
    d2 = {1: [2, 1], 2: [3], '3': [6, 7], 'e': [0]}
    d3 = {1: [1, 2, 2, 1], 2: [3, 4, 3], '3': [5, 6, 7], 'd': [8, 9], 'e': [0]}
    d4 = list_utils.join_key_to_list_dicts(d1, d2)
    assert d4 == d3
    assert d4 != d1


def test_merge_dicts() -> None:
    """Test merge dicts."""
    src = {'a': 5, 'b': 8, 'c': 9}
    dest = {'a': 7, 'c': 4, 'e': 1}
    merged = list_utils.merge_dicts(dest, src)
    assert merged == {'a': 5, 'b': 8, 'c': 9, 'e': 1}
    assert list_utils.merge_dicts(dest, None) == dest
    assert list_utils.merge_dicts(None, None) == {}
    assert list_utils.merge_dicts(None, src) == src
    assert list_utils.merge_dicts({}, {}) == {}


def test_pop_item_from_list() -> None:
    """Test pop item from list."""
    items = [1, 3, 5, 7]
    item = list_utils.pop_item_from_list(items, 5, lambda x: x)
    assert item == 5
    assert items == [1, 3, 7]
    items = []
    assert list_utils.pop_item_from_list(items, 5, lambda x: x) is None
    assert items == []
    assert list_utils.pop_item_from_list(None, None, None) is None
    items = [{'a': 2}, {'b': 4}, {'c': 6}]
    item = list_utils.pop_item_from_list(items, 4, lambda x: list(x.values())[0])
    assert item == {'b': 4}
    assert items == [{'a': 2}, {'c': 6}]


def test_get_item_from_list() -> None:
    """Test get item from list."""
    items = [1, 3, 5, 7]
    item = list_utils.get_item_from_list(items, 5, lambda x: x)
    assert item == 5
    assert items == [1, 3, 5, 7]
    items = []
    assert list_utils.get_item_from_list(items, 5, lambda x: x) is None
    assert items == []
    assert list_utils.get_item_from_list(None, None, None) is None
    items = [{'a': 2}, {'b': 4}, {'c': 6}]
    item = list_utils.get_item_from_list(items, 4, lambda x: list(x.values())[0])
    assert item == {'b': 4}
    assert items == [{'a': 2}, {'b': 4}, {'c': 6}]


def test_comma_sep_to_list() -> None:
    """Test comma sep to list."""
    assert list_utils.comma_sep_to_list(None) == []
    assert list_utils.comma_sep_to_list('a') == ['a']
    assert list_utils.comma_sep_to_list('') == []
    assert list_utils.comma_sep_to_list(' ab , c d,  ef  ') == ['ab', 'c d', 'ef']


def test_comma_colon_sep_to_dict() -> None:
    """Test comma colon sep to dict."""
    assert list_utils.comma_colon_sep_to_dict(None) == {}
    assert list_utils.comma_colon_sep_to_dict('a') == {'a': 'a'}
    assert list_utils.comma_colon_sep_to_dict('') == {}
    assert list_utils.comma_colon_sep_to_dict('  ') == {}
    assert list_utils.comma_colon_sep_to_dict(' ab: cd ,de:fg, h , i : j k, m: n : op ') == {
        'ab': 'cd', 'de': 'fg', 'h': 'h', 'i': 'j k', 'm': 'n : op'
    }


def test_deep_set_get() -> None:
    """Test deep set."""
    d = {}
    path = ['foo', 'bar']
    list_utils.deep_set(d, path, 'hello')
    assert d['foo']['bar'] == 'hello'
    assert list_utils.deep_get(d, path) == 'hello'
    list_utils.deep_set(d, path, 'hi')
    assert d['foo']['bar'] == 'hi'
    assert list_utils.deep_get(d, ['foo', 'car'], 'there') == 'there'
    # set the value to None normally so that it is popped
    list_utils.deep_set(d, path, None)
    assert d['foo'] == {}
    # set it back
    list_utils.deep_set(d, path, 'hi')
    # now set pop_if_none False and the value should be set to None without popping
    list_utils.deep_set(d, path, None, False)
    assert d['foo']['bar'] is None
    path = []
    with pytest.raises(TrestleError):
        list_utils.deep_set(d, path, 'nope')
    with pytest.raises(TrestleError):
        list_utils.deep_get(d, path, 'nope')


def test_deep_update() -> None:
    """Test deep update."""
    d = {}
    path = ['foo', 'bar']
    list_utils.deep_set(d, path, {'a': 'b'})
    assert d['foo']['bar'] == {'a': 'b'}
    list_utils.deep_update(d, path, {'c': 'd'})
    assert d['foo']['bar'] == {'a': 'b', 'c': 'd'}
    with pytest.raises(TrestleError):
        list_utils.deep_update(d, [], {'x': 'y'})
    list_utils.deep_update(d, ['foo'], {'x': 'y'})
    assert d['foo'] == {'bar': {'a': 'b', 'c': 'd'}, 'x': 'y'}



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the ByComponentInterface class."""

from tests import test_utils

import trestle.core.generators as gens
import trestle.oscal.ssp as ossp
from trestle.core.crm.bycomp_interface import ByComponentInterface

test_provided_uuid = '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'
test_responsibility_uuid = '4b34c68f-75fa-4b38-baf0-e50158c13ac2'


def test_get_isolated_responsibilities() -> None:
    """Test retrieving isolated responsibilities statements."""
    by_comp: ossp.ByComponent = test_utils.generate_test_by_comp()
    expected_responsibility = 1
    expected_uuid = by_comp.export.responsibilities[0].uuid  # type: ignore

    bycomp_interface: ByComponentInterface = ByComponentInterface(by_comp)

    result = bycomp_interface.get_isolated_responsibilities()

    assert len(result) == expected_responsibility
    assert result[0].uuid == expected_uuid


def test_get_isolated_provided() -> None:
    """Test retrieving isolated provided statements."""
    by_comp: ossp.ByComponent = test_utils.generate_test_by_comp()
    expected_provided = 1
    expected_uuid = by_comp.export.provided[0].uuid  # type: ignore

    bycomp_interface: ByComponentInterface = ByComponentInterface(by_comp)

    result = bycomp_interface.get_isolated_provided()

    assert len(result) == expected_provided
    assert result[0].uuid == expected_uuid


def test_get_export_sets() -> None:
    """Test retrieving export set statements."""
    by_comp: ossp.ByComponent = test_utils.generate_test_by_comp()
    expected_set = 1
    expected_responsibility_uuid = by_comp.export.responsibilities[1].uuid  # type: ignore
    expected_provided_uuid = by_comp.export.provided[1].uuid  # type: ignore

    bycomp_interface: ByComponentInterface = ByComponentInterface(by_comp)

    result = bycomp_interface.get_export_sets()

    result_set = result[0]

    assert len(result) == expected_set
    assert result_set[0].uuid == expected_responsibility_uuid
    assert result_set[0].provided_uuid == expected_provided_uuid
    assert result_set[1].uuid == expected_provided_uuid


def test_reconcile_inheritance_by_component() -> None:
    """Test retrieving isolated responsibilities statements."""
    by_comp: ossp.ByComponent = test_utils.generate_test_by_comp()
    by_comp.inherited = []
    by_comp.satisfied = []

    # Set up default inherited and satisfied statements
    inherited = gens.generate_sample_model(ossp.Inherited)
    inherited.provided_uuid = test_provided_uuid
    inherited.description = 'inherited description'
    satisfied = gens.generate_sample_model(ossp.Satisfied)
    satisfied.responsibility_uuid = test_responsibility_uuid
    satisfied.description = 'satisfied description'

    by_comp.inherited.append(inherited)
    by_comp.satisfied.append(satisfied)

    bycomp_interface: ByComponentInterface = ByComponentInterface(by_comp)

    # Create new inherited and satisfied statements and update the description
    new_inherited = gens.generate_sample_model(ossp.Inherited)
    new_inherited.provided_uuid = test_provided_uuid
    new_inherited.description = 'new inherited description'
    new_satisfied = gens.generate_sample_model(ossp.Satisfied)
    new_satisfied.responsibility_uuid = test_responsibility_uuid
    new_satisfied.description = 'new satisfied description'

    result_by_comp = bycomp_interface.reconcile_inheritance_by_component([new_inherited], [new_satisfied])

    # Ensure that the resulting by_component has one of each statement and the uuids match the originals
    assert len(result_by_comp.inherited) == 1
    assert len(result_by_comp.satisfied) == 1
    assert result_by_comp.inherited[0].uuid == inherited.uuid  # type: ignore
    assert result_by_comp.satisfied[0].uuid == satisfied.uuid  # type: ignore
    assert result_by_comp.inherited[0].description == new_inherited.description  # type: ignore
    assert result_by_comp.satisfied[0].description == new_satisfied.description  # type: ignore



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handle writing of inherited statements to markdown."""
import logging
import pathlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import trestle.common.const as const
import trestle.core.generators as gens
import trestle.oscal.ssp as ssp
from trestle.common.err import TrestleError
from trestle.core.markdown.docs_markdown_node import DocsMarkdownNode
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.markdown.md_writer import MDWriter

logger = logging.getLogger(__name__)

component_mapping_default: List[Dict[str, str]] = [{'name': const.REPLACE_ME}]


class LeveragedStatements(ABC):
    """Abstract class for managing leveraged statements."""

    def __init__(self, leveraged_ssp_reference: str) -> None:
        """Initialize the class."""
        self._md_file: Optional[MDWriter] = None
        self.header_comment_dict: Dict[str, str] = {
            const.TRESTLE_LEVERAGING_COMP_TAG: const.YAML_LEVERAGING_COMP_COMMENT,
            const.TRESTLE_STATEMENT_TAG: const.YAML_LEVERAGED_COMMENT
        }
        self.merged_header_dict: Dict[str, Any] = {
            const.TRESTLE_STATEMENT_TAG: '',
            const.TRESTLE_LEVERAGING_COMP_TAG: component_mapping_default,
            const.TRESTLE_GLOBAL_TAG: {
                const.LEVERAGED_SSP: {
                    const.HREF: leveraged_ssp_reference
                }
            }
        }

    @abstractmethod
    def write_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Write inheritance information to a single markdown file."""


class StatementTree(LeveragedStatements):
    """Concrete class for managing provided and responsibility statements."""

    def __init__(
        self,
        provided_uuid: str,
        provided_description: str,
        responsibility_uuid: str,
        responsibility_description: str,
        leveraged_ssp_reference: str
    ) -> None:
        """Initialize the class."""
        self.provided_uuid = provided_uuid
        self.provided_description = provided_description
        self.responsibility_uuid = responsibility_uuid
        self.responsibility_description = responsibility_description
        self.satisfied_description = ''
        super().__init__(leveraged_ssp_reference)

    def write_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Write a provided and responsibility statements to a markdown file."""
        self._md_file = MDWriter(leveraged_statement_file, self.header_comment_dict)

        if self._md_file.exists():
            return self.update_statement_md(leveraged_statement_file)

        self._add_generated_content()
        self._md_file.write_out()

    def update_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Update provided and responsibility statements in a markdown file."""
        md_reader = InheritanceMarkdownReader(leveraged_statement_file)

        self.merged_header_dict[const.TRESTLE_LEVERAGING_COMP_TAG] = md_reader.get_leveraged_component_header_value()

        satisfied_description = md_reader.get_satisfied_description()
        if satisfied_description is not None:
            self.satisfied_description = satisfied_description

        self._add_generated_content()
        self._md_file.write_out()

    def _add_generated_content(self) -> None:
        statement_dict: Dict[str, str] = {
            const.PROVIDED_UUID: self.provided_uuid, const.RESPONSIBILITY_UUID: self.responsibility_uuid
        }

        self.merged_header_dict[const.TRESTLE_STATEMENT_TAG] = statement_dict
        self._md_file.add_yaml_header(self.merged_header_dict)

        self._md_file.new_header(level=1, title=const.PROVIDED_STATEMENT_DESCRIPTION)
        self._md_file.new_line(self.provided_description)
        self._md_file.new_header(level=1, title=const.RESPONSIBILITY_STATEMENT_DESCRIPTION)
        self._md_file.new_line(self.responsibility_description)
        self._md_file.new_header(level=1, title=const.SATISFIED_STATEMENT_DESCRIPTION)
        self._md_file.new_line(const.SATISFIED_STATEMENT_COMMENT)
        self._md_file.new_line(self.satisfied_description)


class StatementProvided(LeveragedStatements):
    """Concrete class for managing provided statements."""

    def __init__(
        self,
        provided_uuid: str,
        provided_description: str,
        leveraged_ssp_reference: str,
    ) -> None:
        """Initialize the class."""
        self.provided_uuid = provided_uuid
        self.provided_description = provided_description
        super().__init__(leveraged_ssp_reference)

    def write_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Write provided statements to a markdown file."""
        self._md_file = MDWriter(leveraged_statement_file, self.header_comment_dict)

        if self._md_file.exists():
            return self.update_statement_md(leveraged_statement_file)

        self._add_generated_content()
        self._md_file.write_out()

    def update_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Update provided statements in a markdown file."""
        md_reader = InheritanceMarkdownReader(leveraged_statement_file)

        self.merged_header_dict[const.TRESTLE_LEVERAGING_COMP_TAG] = md_reader.get_leveraged_component_header_value()

        self._add_generated_content()
        self._md_file.write_out()

    def _add_generated_content(self) -> None:
        self.merged_header_dict[const.TRESTLE_STATEMENT_TAG] = {const.PROVIDED_UUID: self.provided_uuid}
        self._md_file.add_yaml_header(self.merged_header_dict)

        self._md_file.new_header(level=1, title=const.PROVIDED_STATEMENT_DESCRIPTION)
        self._md_file.new_line(self.provided_description)


class StatementResponsibility(LeveragedStatements):
    """Concrete class for managing responsibility statements."""

    def __init__(
        self,
        responsibility_uuid: str,
        responsibility_description: str,
        leveraged_ssp_reference: str,
    ) -> None:
        """Initialize the class."""
        self.responsibility_uuid = responsibility_uuid
        self.responsibility_description = responsibility_description
        self.satisfied_description = ''
        super().__init__(leveraged_ssp_reference)

    def write_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Write responsibility statements to a markdown file."""
        self._md_file = MDWriter(leveraged_statement_file, self.header_comment_dict)

        if self._md_file.exists():
            return self.update_statement_md(leveraged_statement_file)

        self._add_generated_content()
        self._md_file.write_out()

    def update_statement_md(self, leveraged_statement_file: pathlib.Path) -> None:
        """Update responsibility statements in a markdown file."""
        md_reader = InheritanceMarkdownReader(leveraged_statement_file)

        self.merged_header_dict[const.TRESTLE_LEVERAGING_COMP_TAG] = md_reader.get_leveraged_component_header_value()

        satisfied_description = md_reader.get_satisfied_description()
        if satisfied_description is not None:
            self.satisfied_description = satisfied_description

        self._add_generated_content()
        self._md_file.write_out()

    def _add_generated_content(self) -> None:
        self.merged_header_dict[const.TRESTLE_STATEMENT_TAG] = {const.RESPONSIBILITY_UUID: self.responsibility_uuid}
        self._md_file.add_yaml_header(self.merged_header_dict)

        self._md_file.new_header(level=1, title=const.RESPONSIBILITY_STATEMENT_DESCRIPTION)
        self._md_file.new_line(self.responsibility_description)
        self._md_file.new_header(level=1, title=const.SATISFIED_STATEMENT_DESCRIPTION)
        self._md_file.new_line(const.SATISFIED_STATEMENT_COMMENT)
        self._md_file.new_line(self.satisfied_description)


@dataclass
class InheritanceInfo:
    """Class to capture component inheritance information."""

    leveraging_comp_titles: List[str]
    inherited: Optional[ssp.Inherited]
    satisfied: Optional[ssp.Satisfied]


class InheritanceMarkdownReader:
    """Class to read leveraged statement information from Markdown."""

    def __init__(self, leveraged_statement_file: pathlib.Path) -> None:
        """Initialize the class."""
        # Save the file name for logging
        self._leveraged_statement_file = leveraged_statement_file

        md_api: MarkdownAPI = MarkdownAPI()

        yaml_header, inheritance_md = md_api.processor.process_markdown(leveraged_statement_file)
        self._yaml_header: Dict[str, Any] = yaml_header
        self._inheritance_md: DocsMarkdownNode = inheritance_md

    def process_leveraged_statement_markdown(self) -> Optional[InheritanceInfo]:
        """
        Read inheritance information from Markdown.

        Returns:
            Optional InheritanceInfo - A list of mapped component titles, an optional satisfied statement and an
            optional inherited statement

        Notes:
            Returns inheritance information in the context of the leveraging SSP. If no leveraging component titles are
            mapped in the yaml header None will be returned. The satisfied and inherited fields are
            generated and returned to added information to by-component assemblies for
            the mapped leveraging components.

        """
        leveraging_comps: List[str] = []
        inherited_statement: Optional[ssp.Inherited] = None
        satisfied_statement: Optional[ssp.Satisfied] = None

        leveraging_comp_header_value: List[Dict[str, str]] = self._yaml_header[const.TRESTLE_LEVERAGING_COMP_TAG]

        # If there are no mapped components, return early
        if not leveraging_comp_header_value or leveraging_comp_header_value == component_mapping_default:
            return None
        else:
            for comp_dicts in leveraging_comp_header_value:
                for comp_title in comp_dicts.values():
                    leveraging_comps.append(comp_title)

        statement_info: Dict[str, str] = self._yaml_header[const.TRESTLE_STATEMENT_TAG]

        if const.PROVIDED_UUID in statement_info:
            # Set inherited

            provided_description = self.get_provided_description()
            if provided_description is None:
                raise TrestleError(f'Provided statement cannot be empty in {self._leveraged_statement_file}')

            inherited_statement = gens.generate_sample_model(ssp.Inherited)

            inherited_statement.description = provided_description
            inherited_statement.provided_uuid = statement_info[const.PROVIDED_UUID]

        if const.RESPONSIBILITY_UUID in statement_info:
            # Set satisfied
            satisfied_description = self.get_satisfied_description()
            if satisfied_description is None:
                raise TrestleError(f'Satisfied statement cannot be empty in {self._leveraged_statement_file}')

            satisfied_statement = gens.generate_sample_model(ssp.Satisfied)

            satisfied_statement.description = satisfied_description
            satisfied_statement.responsibility_uuid = statement_info[const.RESPONSIBILITY_UUID]

        return InheritanceInfo(leveraging_comps, inherited_statement, satisfied_statement)

    def get_leveraged_ssp_reference(self) -> str:
        """Return the leveraged SSP reference in the yaml header."""
        return self._yaml_header[const.TRESTLE_GLOBAL_TAG][const.LEVERAGED_SSP][const.HREF]

    def get_satisfied_description(self) -> Optional[str]:
        """Return the satisfied description in the Markdown."""
        node = self._inheritance_md.get_node_for_key(const.SATISFIED_STATEMENT_DESCRIPTION, False)
        if node is not None:
            return self.strip_heading_and_comments(node.content.raw_text)
        else:
            return None

    def get_provided_description(self) -> Optional[str]:
        """Return the provided description in the Markdown."""
        node = self._inheritance_md.get_node_for_key(const.PROVIDED_STATEMENT_DESCRIPTION, False)
        if node is not None:
            return self.strip_heading_and_comments(node.content.raw_text)
        else:
            return None

    def get_leveraged_component_header_value(self) -> Dict[str, str]:
        """Provide the leveraged component value in the yaml header."""
        return self._yaml_header[const.TRESTLE_LEVERAGING_COMP_TAG]

    @staticmethod
    def strip_heading_and_comments(markdown_text: str) -> str:
        """Remove the heading and comments from lines to get the multi-line paragraph."""
        lines = markdown_text.split('\n')
        non_heading_comment_lines = []
        inside_heading = False
        inside_comment = False

        for line in lines:
            if line.startswith('#'):
                inside_heading = True
            elif line.startswith('<!--'):
                inside_comment = True
                if line.rstrip().endswith('-->'):
                    inside_comment = False
                continue
            elif line.strip() == '':
                inside_heading = False
                inside_comment = False

            if not inside_heading and not inside_comment:
                non_heading_comment_lines.append(line)

        stripped_markdown = '\n'.join(non_heading_comment_lines).strip()
        return stripped_markdown



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Automation for documentation building.

mkdocs is a great system, however, there are a few areas where it's a bit messy.

this script is designed to close a few automation gaps for files that it is desirable
they are included and must be moved.

::: trestle.oscal.catalog
handler: python
"""
import pathlib
import shutil
from typing import Any, Dict, List

from ruamel.yaml import YAML


def update_mkdocs_meta(path: pathlib.Path, module_list: List[Any]) -> None:
    """Update the mkdocs.yml structure file to represent the latest trestle modules."""
    yaml = YAML()
    fh_read = path.open('r', encoding='utf8')
    yaml_structure = yaml.load(fh_read)
    fh_read.close()
    nav = yaml_structure['nav']
    for index in range(len(nav)):
        if 'Reference' in nav[index].keys():
            nav[index]['Reference'] = []
            nav[index]['Reference'].append({'Integrating with IBM SCC': 'reference/third-party-result-schema-SCC.md'})
            nav[index]['Reference'].append({'trestle API reference': module_list})

    yaml_structure['nav'] = nav
    fh_write = path.open('w', encoding='utf8')
    yaml.dump(yaml_structure, fh_write)
    fh_write.close()


def write_module_doc_metafile(dump_location: pathlib.Path, module_name: str) -> None:
    """Create a markdown file which allows mkdocstrings to index an individual module."""
    write_file = dump_location / (module_name + '.md')
    fh = write_file.open('w', encoding='utf8')
    fh.write(f'::: {module_name}\n')
    fh.write('handler: python\n')
    fh.close()


def get_key(var: Dict[str, Any]) -> str:
    """Get the first key from a dict."""
    # Dict has only one key
    return list(var.keys())[0]


def create_module_markdowns(base_path: pathlib.Path, base_module: str, dump_location: pathlib.Path) -> List[Any]:
    """Create markdown files for running mkdocstrings and return the desired structure for use in the ToC."""
    # First clean the directory
    module_arr = []
    for path in base_path.iterdir():
        module_full_name = base_module + '.' + path.stem
        struct = None
        if path.is_dir() and path.name[0] != '_':
            struct = create_module_markdowns(path, module_full_name, dump_location)
            if len(struct) > 0:
                module_arr.append({path.stem: struct})
        elif path.suffix == '.py' and path.stem[0] != '_':
            struct = dump_location.stem + '/' + module_full_name + '.md'
            write_module_doc_metafile(dump_location, module_full_name)
            module_arr.append({path.stem: struct})
    module_arr.sort(key=get_key)
    return module_arr


def md_txt(original: pathlib.Path, md_ed_license: pathlib.Path):
    """Convert the apache license into a markdown file by wrapping the content in a text escape block."""
    assert original.exists()
    orig_data = original.open('r', encoding='utf8').read()
    md_fh = md_ed_license.open('w', encoding='utf8')
    md_fh.write('```text\n')
    md_fh.write(orig_data)
    md_fh.write('\n```\n')


def cleanup_directory(dump_location: pathlib.Path):
    """Cleanup a directory by removing all content and recreating."""
    if dump_location.is_dir():
        shutil.rmtree(dump_location)
        dump_location.mkdir()
    elif dump_location.is_file():
        raise Exception(f'{dump_location} Cannot be a file')
    else:
        dump_location.mkdir()


if __name__ == '__main__':
    # Setup structure automatically for mkdocstrings
    api_ref_location = pathlib.Path('docs/api_reference')
    cleanup_directory(api_ref_location)
    structer = create_module_markdowns(pathlib.Path('trestle'), 'trestle', api_ref_location)
    update_mkdocs_meta(pathlib.Path('mkdocs.yml'), structer)
    # Ensure single source of truth for license file
    md_txt(pathlib.Path('LICENSE'), pathlib.Path('docs/license.md'))
    md_txt(pathlib.Path('DCO1.1.txt'), pathlib.Path('docs/contributing/DCO.md'))



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handle reading of writing controls from markdown."""
import logging
import pathlib
from typing import Any, Dict, List, Optional, Tuple

import trestle.core.generic_oscal as generic
import trestle.oscal.catalog as cat
from trestle.common import const
from trestle.common.common_types import TypeWithProps
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_list, deep_get, delete_list_from_list, none_if_empty
from trestle.common.model_utils import ModelUtils
from trestle.core import generators as gens
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_interface import CompDict, ComponentImpInfo, ControlInterface
from trestle.core.markdown.control_markdown_node import ControlMarkdownNode
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.oscal import common
from trestle.oscal import component as comp
from trestle.oscal import profile as prof
from trestle.oscal import ssp as ossp

logger = logging.getLogger(__name__)


class ControlReader:
    """Class to read controls from markdown."""

    @staticmethod
    def _clean_prose(prose: List[str]) -> List[str]:
        # remove empty and horizontal rule lines at start and end of list of prose lines
        forward_index = 0
        for line in prose:
            if line.strip() and not line.startswith('____'):
                break
            forward_index += 1
        new_prose = prose[forward_index:]
        reverse_index = 0
        for line in reversed(new_prose):
            if line.strip() and not line.startswith('____'):
                break
            reverse_index += 1
        clean_prose = new_prose[:len(new_prose) - reverse_index]
        clean_prose = clean_prose if clean_prose else ['']
        # if there is no useful prose this will return [''] and allow generation of a statement with empty prose
        return clean_prose

    @staticmethod
    def _comp_name_in_dict(comp_name: str, comp_dict: CompDict) -> str:
        """If the name is already in the dict in a similar form, stick to that form."""
        simple_name = ControlReader.simplify_name(comp_name)
        for name in comp_dict.keys():
            if simple_name == ControlReader.simplify_name(name):
                return name
        return comp_name

    @staticmethod
    def _add_node_to_dict(
        comp_name: Optional[str],
        label: str,
        comp_dict: CompDict,
        node: ControlMarkdownNode,
        control_id: str,
        comp_list: List[str],
        context: ControlContext
    ) -> None:
        """
        Extract the label, prose, possible component name - along with implementation status.

        In component mode there is only one component and its name is not in markdown.
        In ssp mode there are many components in each md file.
        """
        component_mode = context.purpose == ContextPurpose.COMPONENT
        # for ssp, ### marks component name but for component it is ##
        # if it is a header, make sure it has correct format
        if node.key and node.key[0] == '#' and ControlInterface.bad_header(node.key):
            raise TrestleError(f'Improper header format for control {control_id}: {node.key}')
        if not component_mode:
            # look for component name heading if present
            prefix = '### '
            if node.key.startswith(prefix):
                if len(node.key.split()) <= 1:
                    raise TrestleError(
                        f'Header line in control {control_id} markdown starts with {prefix} but has no content.'
                    )
                comp_name = node.key.split(' ', 1)[1].strip()
                simp_comp_name = ControlReader.simplify_name(comp_name)
                if simp_comp_name in comp_list:
                    raise TrestleError(
                        f'Control {control_id} has a section with two component headings for {comp_name}.  '
                        'Please combine the sections so there is only one heading for each component in a '
                        'statement.'
                    )
                comp_list.append(simp_comp_name)
                comp_name = ControlReader._comp_name_in_dict(comp_name, comp_dict)

        # prose may be empty in md and we want to capture that so put it in the comp_dict
        prose = '\n'.join(ControlReader._clean_prose(node.content.text))
        # add the prose to the comp_dict, creating new entry as needed
        if comp_name in comp_dict:
            if label in comp_dict[comp_name]:
                comp_dict[comp_name][label].prose = prose
            else:
                # create new entry with prose
                comp_dict[comp_name][label] = ComponentImpInfo(prose=prose, rules=[], props=[])
        elif comp_name:
            comp_dict[comp_name] = {label: ComponentImpInfo(prose=prose, rules=[], props=[])}

        # build list of subnodes that get handled specially so they aren't processed here
        subnode_kill: List[int] = []
        status_str = None
        remarks_str = None
        rules_list: List[str] = []
        for ii, subnode in enumerate(node.subnodes):
            if subnode.key.find(const.IMPLEMENTATION_STATUS_REMARKS_HEADER) >= 0:
                remarks_str = subnode.key.split(maxsplit=4)[-1]
                subnode_kill.append(ii)
            elif subnode.key.find(const.IMPLEMENTATION_STATUS_HEADER) >= 0:
                status_str = subnode.key.split(maxsplit=3)[-1]
                subnode_kill.append(ii)
            elif subnode.key.find('Rules:') >= 0:
                rules_list = [text[2:] for text in subnode.content.text if text.startswith('- ')]
                subnode_kill.append(ii)
        if status_str:
            new_status = common.ImplementationStatus(state=status_str, remarks=remarks_str)
            if comp_name not in comp_dict:
                comp_dict[comp_name] = {}
            if label not in comp_dict[comp_name]:
                comp_dict[comp_name][label] = ComponentImpInfo(prose='', rules=[], props=[])
            comp_dict[comp_name][label].status = new_status
        if rules_list:
            comp_dict[comp_name][label].rules = rules_list
        delete_list_from_list(node.subnodes, subnode_kill)
        for subnode in as_list(node.subnodes):
            ControlReader._add_node_to_dict(comp_name, label, comp_dict, subnode, control_id, comp_list, context)

    @staticmethod
    def _insert_header_content(
        imp_req: generic.GenericImplementedRequirement, header: Dict[str, Any], control_id: str
    ) -> None:
        """Insert yaml header content into the imp_req and its by_comps as props."""
        dict_ = header.get(const.TRESTLE_PROPS_TAG, {})
        roles = as_list(dict_.get(const.RESPONSIBLE_ROLES, []))
        props = []
        responsible_roles = []
        for role in roles:
            if isinstance(role, str):
                # role_id must conform to NCNAME regex
                role = role.strip().replace(' ', '_')
                if role:
                    responsible_roles.append(common.ResponsibleRole(role_id=role))
            else:
                logger.warning(f'Role in header for control {control_id} not recognized: {role}')
        if props:
            imp_req.props = as_list(imp_req.props)
            imp_req.props.extend(props)
        if responsible_roles:
            imp_req.responsible_roles = as_list(imp_req.responsible_roles)
            imp_req.responsible_roles.extend(responsible_roles)
            imp_req.responsible_roles = none_if_empty(imp_req.responsible_roles)
            # enforce single list of resp. roles for control and each by_comp
            for by_comp in as_list(imp_req.by_components):
                by_comp.responsible_roles = imp_req.responsible_roles

    @staticmethod
    def simplify_name(name: str) -> str:
        """Simplify the name to ignore variations in case, space, hyphen, underscore, slash."""
        return name.lower().replace(' ', '').replace('-', '').replace('_', '').replace('/', '')

    @staticmethod
    def _get_label_from_implementation_header(imp_header: str):
        # assumed to be of form: Implementation for part a.
        split_header = imp_header.split(' ', 4)
        if len(split_header) != 5:
            raise TrestleError(f'Implementation header cannot be parsed for statement part: {imp_header}')
        return split_header[4].strip()

    @staticmethod
    def read_control_info_from_md(control_file: pathlib.Path,
                                  context: ControlContext) -> Tuple[Dict[str, List[str]], CompDict]:
        """
        Find all labels and associated implementation prose in the markdown for this control.

        Args:
            control_file: path to the control markdown file
            context: context of the control usage

        Returns:
            The yaml header as dict in second part of tuple.
            Adds to the passed in comp_dict.
        """
        yaml_header = {}
        comp_dict = {}

        if not control_file.exists():
            return yaml_header, comp_dict
        # if the file exists, load the contents but do not use prose from comp_dict
        # for non ssp or component mode just use empty string for comp
        comp_name = ''
        try:
            control_id = control_file.stem
            if context.purpose == ContextPurpose.COMPONENT:
                comp_name = context.comp_name if context.comp_name else const.SSP_MAIN_COMP_NAME

            md_api = MarkdownAPI()
            yaml_header, control_md = md_api.processor.process_markdown(control_file)

            # first get the header strings, including statement labels, for statement imp reqs
            imp_string = '## Implementation '
            headers = control_md.get_all_headers_for_level(2)
            # get e.g. ## Implementation a.  ## Implementation b. etc
            imp_header_list = [header for header in headers if header.startswith(imp_string)]

            # now get the (one) header for the main solution
            main_headers = list(control_md.get_all_headers_for_key(const.SSP_MD_IMPLEMENTATION_QUESTION, False))
            # should be only one header, so warn if others found
            if main_headers:
                if len(main_headers) > 1:
                    logger.warning(
                        f'Control {control_id} has {len(main_headers)} main header responses.  Will use first one only.'
                    )
                main_header = main_headers[0]
                node = control_md.get_all_nodes_for_keys([main_header], False)[0]
                # this node is top level so it will have empty label
                # it may have subnodes of Rules, Implementation Status, Implementaton Remarks
                ControlReader._add_node_to_dict(comp_name, '', comp_dict, node, control_id, [], context)
            for imp_header in imp_header_list:
                label = ControlReader._get_label_from_implementation_header(imp_header)
                node = control_md.get_node_for_key(imp_header)
                ControlReader._add_node_to_dict(comp_name, label, comp_dict, node, control_id, [], context)

        except TrestleError as e:
            raise TrestleError(f'Error occurred reading {control_file}: {e}')
        return yaml_header, comp_dict

    @staticmethod
    def _handle_empty_prose(prose: str, id_: str) -> str:
        """Regard prompt text or id_ as no prose and return blank string."""
        if prose.startswith(const.SSP_ADD_IMPLEMENTATION_PREFIX) or prose == id_:
            return ''
        return prose

    @staticmethod
    def read_implemented_requirement(control_file: pathlib.Path,
                                     context: ControlContext) -> Tuple[str, comp.ImplementedRequirement]:
        """
        Get the implementated requirement associated with given control and link to existing components or new ones.

        Args:
            control_file: path of the control markdown file
            context: context of the control usage

        Returns:
            Tuple: The control sort-id and the one implemented requirement for this control.

        Notes:
            Each statement may have several responses, with each response in a by_component for a specific component.
            statement_map keeps track of statements that may have several by_component responses.
            This is only used during component assemble and only for updating one component.
        """
        control_id = control_file.stem
        md_header, md_comp_dict = ControlReader.read_control_info_from_md(control_file, context)
        comp_name = context.component.title

        statement_map: Dict[str, comp.Statement] = {}
        # create a new implemented requirement linked to the control id to hold the statements
        imp_req = gens.generate_sample_model(comp.ImplementedRequirement)
        imp_req.control_id = control_id

        imp_req.statements = []
        comp_dict = md_comp_dict[comp_name]
        for label, comp_info in comp_dict.items():
            # only assemble responses with associated rules
            if not comp_info.rules:
                continue
            # if no label it applies to the imp_req itself rather than a statement
            if not label:
                imp_req.description = ControlReader._handle_empty_prose(comp_info.prose, control_id)
                ControlInterface.insert_status_in_props(imp_req, comp_info.status)
                continue
            statement_id = ControlInterface.create_statement_id(control_id)
            if label in ['', const.STATEMENT]:
                statement_part_id = statement_id
            else:
                clean_label = label.strip('.')
                statement_part_id = ControlInterface.strip_to_make_ncname(f'{statement_id}.{clean_label}')
            if statement_part_id in statement_map:
                statement = statement_map[statement_part_id]
            else:
                statement = gens.generate_sample_model(comp.Statement)
                statement.statement_id = statement_part_id
                statement_map[statement_part_id] = statement
            statement.description = comp_info.prose
            statement.props = none_if_empty(ControlInterface.clean_props(comp_info.props))
            ControlInterface.insert_status_in_props(statement, comp_info.status)

        imp_req.statements = list(statement_map.values())
        imp_req.set_parameters = []

        for _, param_dict_list in md_header.get(const.COMP_DEF_RULES_PARAM_VALS_TAG, {}).items():
            for param_dict in param_dict_list:
                values = param_dict.get(const.VALUES, [])
                comp_values = param_dict.get(const.COMPONENT_VALUES, [])
                values = comp_values if comp_values else values
                set_param = ossp.SetParameter(param_id=param_dict['name'], values=values)
                imp_req.set_parameters.append(set_param)
        imp_req.statements = none_if_empty(list(statement_map.values()))
        imp_req.set_parameters = none_if_empty(imp_req.set_parameters)

        ControlReader._insert_header_content(imp_req, md_header, control_id)
        sort_id = md_header.get(const.SORT_ID, control_id)
        return sort_id, imp_req

    @staticmethod
    def _get_props_list(control_id: str, label_map: Dict[str, str],
                        yaml_header: Dict[str, Any]) -> Tuple[List[common.Property], Dict[str, List[common.Property]]]:
        """Get the list of props in the yaml header of this control as separate lists with and without by_id."""
        prop_list = yaml_header.get(const.TRESTLE_ADD_PROPS_TAG, [])
        props = []
        props_by_id = {}
        for prop_d in prop_list:
            by_id = prop_d.get('smt-part', None)
            if by_id and control_id in label_map:
                by_id = label_map[control_id].get(by_id, by_id)
            prop = common.Property(name=prop_d['name'], value=prop_d['value'], ns=prop_d.get('ns', None))
            if by_id:
                if by_id not in props_by_id:
                    props_by_id[by_id] = []
                props_by_id[by_id].append(prop)
            else:
                props.append(prop)
        return props, props_by_id

    @staticmethod
    def read_editable_content(
        control_path: pathlib.Path,
        required_sections_list: List[str],
        part_label_to_id_map: Dict[str, Dict[str, str]],
        cli_section_dict: Dict[str, str],
        write_mode: bool
    ) -> Tuple[str, List[prof.Alter], Dict[str, Any]]:
        """Get parts for the markdown control corresponding to Editable Content - along with the set-parameter dict."""
        control_id = control_path.stem

        md_api = MarkdownAPI()
        yaml_header, control_tree = md_api.processor.process_control_markdown(control_path, cli_section_dict, part_label_to_id_map)  # noqa: E501
        # extract the sort_id if present in header
        sort_id = yaml_header.get(const.SORT_ID, control_id)

        editable_node = None
        for header in list(control_tree.get_all_headers_for_level(1)):
            if header.startswith('# Editable'):
                editable_node = control_tree.get_node_for_key(header)
                break
        if not editable_node:
            return sort_id, [], {}

        editable_parts = control_tree.get_editable_parts_and_subparts()
        by_id_parts = control_tree.get_by_id_parts()
        found_sections = [p.name for p in editable_parts]

        # Validate that all required sections have a prose
        for editable_part in editable_parts:
            if not write_mode and editable_part.name in required_sections_list and editable_part.prose.startswith(
                    const.PROFILE_ADD_REQUIRED_SECTION_FOR_CONTROL_TEXT):
                raise TrestleError(f'Control {control_id} is missing prose for required section {editable_part.title}')

        # Validate that all required sections are present
        missing_sections = set(required_sections_list) - set(found_sections)
        if missing_sections:
            raise TrestleError(f'Control {control_id} is missing required sections {missing_sections}')
        param_dict: Dict[str, Any] = {}
        # get set_params from the header and add to parm_dict
        header_params = yaml_header.get(const.SET_PARAMS_TAG, {})
        if header_params:
            param_dict.update(header_params)

        props, props_by_id = ControlReader._get_props_list(control_id, part_label_to_id_map, yaml_header)

        # When adding props without by_id it can either be starting or ending and we default to ending
        # This is the default behavior as described for implicit binding in
        # https://pages.nist.gov/OSCAL/concepts/processing/profile-resolution/
        # When adding props to a part using by_id, it is the same situation because it cannot be before or after since
        # props are not in the same list as parts

        adds: List[prof.Add] = []

        # add the parts and props at control level
        if editable_parts or props:
            adds.append(
                prof.Add(
                    parts=none_if_empty(editable_parts), props=none_if_empty(props), position=prof.Position.ending
                )
            )

        # add the parts and props at the part level, by-id
        by_ids = set(by_id_parts.keys()).union(props_by_id.keys())
        for by_id in sorted(by_ids):
            parts = by_id_parts.get(by_id, None)
            props = props_by_id.get(by_id, None)
            adds.append(prof.Add(parts=parts, props=props, position=prof.Position.ending, by_id=by_id))

        new_alters = []
        if adds:
            new_alters = [prof.Alter(control_id=control_id, adds=adds)]
        return sort_id, new_alters, param_dict

    @staticmethod
    def _update_display_prop_namespace(item: TypeWithProps):
        """Set namespace for special property display_name."""
        for prop in as_list(item.props):
            if prop.name == const.DISPLAY_NAME:
                prop.ns = const.TRESTLE_GENERIC_NS

    @staticmethod
    def read_control(control_path: pathlib.Path, set_parameters_flag: bool) -> Tuple[cat.Control, str]:
        """Read the control and group title from the markdown file."""
        control = gens.generate_sample_model(cat.Control)
        md_api = MarkdownAPI()
        yaml_header, control_tree = md_api.processor.process_control_markdown(control_path)
        control_titles = list(control_tree.get_all_headers_for_level(1))
        if len(control_titles) == 0:
            raise TrestleError(f'Control markdown: {control_path} contains no control title.')
        if len(control_titles) > 1:
            raise TrestleError(f'Control markdown: {control_path} contains multiple control titles {control_titles}.')

        control.id = control_tree.subnodes[0].content.control_id
        group_title = control_tree.subnodes[0].content.control_group
        control.title = control_tree.subnodes[0].content.control_title

        control_statement = control_tree.get_control_statement()
        statement_part = control_statement.content.part

        control.parts = [statement_part] if statement_part else None
        control_objective = control_tree.get_control_objective()
        if control_objective is not None:
            objective_part = control_objective.content.part
            if objective_part:
                if control.parts:
                    control.parts.append(objective_part)
                else:
                    control.parts = [objective_part]

        control_guidance = control_tree.get_control_guidance()
        if control_guidance is not None:
            guidance_part = control_guidance.content.part
            if guidance_part:
                if control.parts:
                    control.parts.append(guidance_part)
                else:
                    control.parts = [guidance_part]

        all_other_parts = []
        for section_node in control_tree.get_other_control_parts():
            parts = section_node.content.part
            all_other_parts.extend([parts])
        if all_other_parts:
            if control.parts:
                control.parts.extend(all_other_parts)
            else:
                control.parts = all_other_parts

        if set_parameters_flag:
            params: Dict[str, str] = yaml_header.get(const.SET_PARAMS_TAG, [])
            if params:
                control.params = []
                for id_, param_dict in params.items():
                    param_dict['id'] = id_
                    param = ModelUtils.dict_to_parameter(param_dict)
                    # if display_name is in list of properties, set its namespace
                    ControlReader._update_display_prop_namespace(param)
                    control.params.append(param)
        sort_id = deep_get(yaml_header, [const.TRESTLE_GLOBAL_TAG, const.SORT_ID], None)
        if sort_id:
            control.props = control.props if control.props else []
            control.props.append(common.Property(name=const.SORT_ID, value=sort_id))
        return control, group_title



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL transformation tasks."""

# mypy: ignore-errors  # noqa E800
import configparser
import datetime
import json
import logging
import os
import pathlib
import traceback
import uuid
from typing import Dict, List, Optional, Tuple

import trestle
from trestle.common import const
from trestle.oscal import OSCAL_VERSION
from trestle.oscal.common import Metadata
from trestle.oscal.common import Party
from trestle.oscal.common import Property
from trestle.oscal.common import ResponsibleParty
from trestle.oscal.common import ResponsibleRole
from trestle.oscal.common import Role
from trestle.oscal.component import ComponentDefinition
from trestle.oscal.component import ControlImplementation
from trestle.oscal.component import DefinedComponent
from trestle.oscal.component import ImplementedRequirement
from trestle.oscal.component import SetParameter
from trestle.tasks.base_task import TaskBase
from trestle.tasks.base_task import TaskOutcome

logger = logging.getLogger(__name__)


class Ocp4CisProfileToOscalCD(TaskBase):
    """
    Task to transform OCP4 CIS profile to OSCAL component-definition.

    Attributes:
        name: Name of the task.
    """

    name = 'ocp4-cis-profile-to-oscal-cd'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task ocp4-cis-profile-to-oscal-cd.

        Args:
            config_object: Config section associated with the task.
        """
        super().__init__(config_object)
        self._timestamp = datetime.datetime.utcnow().replace(microsecond=0).replace(tzinfo=datetime.timezone.utc
                                                                                    ).isoformat()

    def set_timestamp(self, timestamp: str) -> None:
        """Set the timestamp."""
        self._timestamp = timestamp

    def print_info(self) -> None:
        """Print the help string."""
        logger.info(f'Help information for {self.name} task.')
        logger.info('')
        logger.info('Purpose: Create component definition from standard (e.g. CIS benchmark).')
        logger.info('')
        logger.info('Configuration flags sit under [task.ocp4-cis-profile-to-oscal-cd]:')
        text1 = '  component-name         = '
        text2 = 'component name, e.g. OSCO.'
        logger.info(text1 + text2)
        text1 = '  org-name               = '
        text2 = 'organization name, e.g. International Business Machines.'
        logger.info(text1 + text2)
        text1 = '  org-remarks            = '
        text2 = 'organization remarks, e.g. IBM.'
        logger.info(text1 + text2)
        text1 = '  folder-cac             = '
        text2 = 'folder containing compliance-as-code artifacts, e.g adjunct-data/cis-benchmarks/content.'
        logger.info(text1 + text2)
        text1 = '  output-dir             = '
        text2 = 'location to write the generated component-definition.json file.'
        logger.info(text1 + text2)
        #
        text1 = '  profile-name           = '
        text2 = 'profile name, e.g. OCP4 CIS-benchmark v4.'
        logger.info(text1 + text2)
        text1 = '  profile-mnemonic       = '
        text2 = 'profile mnemonic, e.g. ocp4-cis-node.'
        logger.info(text1 + text2)
        text1 = '  profile-ns             = '
        text2 = 'profile ns, e.g. https://oscal-compass.github.io/compliance-trestle/schemas/oscal/ibm-cloud.'
        logger.info(text1 + text2)
        text1 = '  profile-version        = '
        text2 = 'profile version, e.g. 1.1.'
        logger.info(text1 + text2)
        text1 = '  profile-check-version  = '
        text2 = 'profile check version, e.g. 0.1.58.'
        logger.info(text1 + text2)
        #
        text1 = '  profile-type           = '
        text2 = 'profile type, e.g. OCP4.'
        logger.info(text1 + text2)
        text1 = '  profile-list           = '
        text2 = 'profile list is blank separated list of "<suffix>" for config entries: '
        logger.info(text1 + text2)
        text1 = '                           profile-file.<suffix>, profile-title.<suffix>, profile-url.<suffix>'
        text2 = ', e.g. cis cis-node.'
        logger.info(text1 + text2)
        text1 = '  profile-file.<suffix>  = '
        text2 = 'path of the profile file to ingest'
        text3 = ', e.g. ${folder-cac}/products/ocp4/profiles/cis-node.profile.'  # noqa
        logger.info(text1 + text2 + text3)
        text1 = '  profile-title.<suffix> = '
        text2 = 'title of the profile'
        text3 = ', e.g. CIS Red Hat OpenShift Container Platform 4 Benchmark.'
        logger.info(text1 + text2 + text3)
        text1 = '  profile-url.<suffix>   = '
        text2 = 'URL of the profile'
        text3 = ', e.g. https://github.com/ComplianceAsCode/content/blob/master/products/ocp4/profiles/cis.profile.'
        logger.info(text1 + text2 + text3)
        text1 = '  rule-to-parameters-map = '
        text2 = 'map file for set-parameters, e.g. '
        text3 = 'adjunct-data/task-files/rule2var.json.'
        logger.info(text1 + text2 + text3)
        text1 = '  selected-rules         = '
        text2 = 'file with list of selected rules, e.g. '
        text3 = 'adjunct-data/task-files/selected_rules.json.'
        logger.info(text1 + text2 + text3)
        text1 = '  enabled-rules          = '
        text2 = 'file with list of enabled rules, e.g. '
        text3 = 'adjunct-data/task-files/enabled_rules.json.'
        logger.info(text1 + text2 + text3)
        #
        text = ''
        logger.info(text)
        text = 'Notes:'
        logger.info(text)
        text = '1. If a control has selected rules but no enabled rules, then all those selected are included.'
        logger.info(text)
        text = '2. If a control has selected and enabled rules, then only those enabled are included.'
        logger.info(text)
        text = '3. If a control has no selected rules, then none are included regardless of enabled.'
        logger.info(text)

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        return TaskOutcome('simulated-success')

    def execute(self) -> TaskOutcome:
        """Provide an actual outcome."""
        try:
            return self._execute()
        except Exception:
            logger.info(traceback.format_exc())
            return TaskOutcome('failure')

    def _execute(self) -> TaskOutcome:
        if not self._config:
            logger.warning('config missing')
            return TaskOutcome('failure')
        try:
            component_name = self._config['component-name']
            org_name = self._config['org-name']
            org_remarks = self._config['org-remarks']
            self._folder_cac = self._config['folder-cac']
            profile_check_version = self._config['profile-check-version']
            profile_type = self._config['profile-type']
            profile_mnemonic = self._config['profile-mnemonic']
            profile_name = self._config['profile-name']
            profile_ns = self._config['profile-ns']
            profile_version = self._config['profile-version']
            profile_sets = {}
            profile_list = self._config['profile-list'].split()
            for profile in profile_list:
                profile_sets[profile] = {}
                profile_sets[profile]['profile-file'] = self._config[f'profile-file.{profile}']
                profile_sets[profile]['profile-url'] = self._config[f'profile-url.{profile}']
                profile_sets[profile]['profile-title'] = self._config[f'profile-title.{profile}']
                profile_sets[profile]['profile-ns'] = profile_ns
                profile_sets[profile]['component-name'] = component_name
            odir = self._config['output-dir']
        except KeyError as e:
            logger.info(f'key {e.args[0]} missing')
            return TaskOutcome('failure')
        # selected rules
        self._selected_rules = self._get_filter_rules('selected-rules', 'selected')
        # enabled rules
        self._enabled_rules = self._get_filter_rules('enabled-rules', 'enabled')
        # verbosity
        quiet = self._config.get('quiet', False)
        verbose = not quiet
        # output
        overwrite = self._config.getboolean('output-overwrite', True)
        opth = pathlib.Path(odir)
        # insure output dir exists
        opth.mkdir(exist_ok=True, parents=True)
        # calculate output file name & check writability
        oname = 'component-definition.json'
        ofile = opth / oname
        if not overwrite and pathlib.Path(ofile).exists():
            logger.warning(f'output: {ofile} already exists')
            return TaskOutcome('failure')
        # fetch rule to parameters map
        self._rule_to_parm_map = self._get_parameters_map('rule-to-parameters-map')
        # roles, responsible_roles, parties, responsible parties
        party_uuid_01 = str(uuid.uuid4())
        party_uuid_02 = str(uuid.uuid4())
        party_uuid_03 = str(uuid.uuid4())
        roles = self._build_roles()
        responsible_roles = self._build_responsible_roles(party_uuid_01, party_uuid_02, party_uuid_03)
        parties = self._build_parties(org_name, org_remarks, party_uuid_01, party_uuid_02, party_uuid_03)
        responsible_parties = self._build_responsible_parties(party_uuid_01, party_uuid_02, party_uuid_03)
        # metadata
        metadata = Metadata(
            title=f'Component definition for {profile_type} profiles',
            last_modified=self._timestamp,
            oscal_version=OSCAL_VERSION,
            version=trestle.__version__,
            roles=roles,
            parties=parties,
            responsible_parties=responsible_parties
        )
        # defined component
        component_title = component_name
        component_description = component_name
        defined_component = DefinedComponent(
            uuid=str(uuid.uuid4()),
            description=component_description,
            title=component_title,
            type='Service',
        )
        # add control implementation per profile
        prop1 = Property(
            name='profile_name',
            value=profile_name,
            class_='scc_profile_name',
            ns=profile_ns,
        )
        prop2 = Property(
            name='profile_mnemonic',
            value=profile_mnemonic,
            class_='scc_profile_mnemonic',
            ns=profile_ns,
        )
        prop3 = Property(
            name='profile_version',
            value=profile_version,
            class_='scc_profile_version',
            ns=profile_ns,
        )
        prop4 = Property(
            name='profile_check_version',
            value=profile_check_version,
        )
        props = [prop1, prop2, prop3, prop4]
        for profile in profile_list:
            profile_set = profile_sets[profile]
            control_implementation = self._build_control_implementation(profile_set, responsible_roles, props)
            if control_implementation is not None:
                if defined_component.control_implementations is None:
                    defined_component.control_implementations = [control_implementation]
                else:
                    defined_component.control_implementations.append(control_implementation)
        # defined components
        defined_components = [defined_component]
        # component definition
        component_definition = ComponentDefinition(
            uuid=str(uuid.uuid4()),
            metadata=metadata,
            components=defined_components,
        )
        # write OSCAL ComponentDefinition to file
        if verbose:
            logger.info(f'output: {ofile}')
        component_definition.oscal_write(pathlib.Path(ofile))
        return TaskOutcome('success')

    def _get_set_parameter(self, rule: str) -> SetParameter:
        """Get set parameter."""
        set_parameter = None
        for key in self._rule_to_parm_map.keys():
            logger.debug(f'{key} {rule}')
            if key == rule:
                value = self._rule_to_parm_map[key]
                remarks = value['description']
                options = value['options']
                default_value = options['default']
                logger.debug(f'key: {key} options: {options}')
                set_parameter = SetParameter(
                    param_id=rule,
                    values=[f'{default_value}'],
                    remarks=remarks,
                )
        return set_parameter

    def _get_controls(self, rules: Dict[str, Tuple[str, str, str]]) -> Dict[str, List[str]]:
        """Get controls."""
        controls = {}
        for rule in rules.keys():
            control = rules[rule][1]
            if control not in controls.keys():
                controls[control] = [rule]
            else:
                controls[control] = controls[control] + [rule]
        # trim rules associated with control with respect to enabled rules
        for control in controls:
            controls[control] = self._get_trimmed_rules(control, controls[control])
            logger.debug(f'{control} {controls[control]}')
        return controls

    # determine if trim is needed for the control, and if so then
    # for the associated set of rules drop those that are not enabled
    def _get_trimmed_rules(self, control: str, rules_for_control: List[str]) -> List[str]:
        """Trim rules if any rule for control appears in enabled rules list."""
        retval = rules_for_control
        if self._is_trim_needed(rules_for_control):
            retval = []
            for rule in rules_for_control:
                if rule in self._enabled_rules:
                    retval = retval + [rule]
                    logger.debug(f'keep {control} {rule}')
                else:
                    logger.debug(f'drop {control} {rule}')
        return retval

    # if any rule in the set of rules for the control appears in the enabled list,
    # then trim is needed
    def _is_trim_needed(self, rules_for_control: List[str]) -> bool:
        """Check if trim is needed."""
        retval = False
        for rule in rules_for_control:
            if rule in self._enabled_rules:
                retval = True
                break
        return retval

    # fetch the set of rules that will be included/excluded from the CIS rules
    def _get_parameters_map(self, config_key: str) -> List[str]:
        """Get parameters map."""
        try:
            fp = pathlib.Path(self._config[config_key])
            f = fp.open('r', encoding=const.FILE_ENCODING)
            jdata = json.load(f)
            parameters_map = jdata
            f.close()
        except KeyError as e:
            logger.debug(f'key {e.args[0]} missing')
            parameters_map = {}
        except Exception:
            logger.warning(f'unable to process {self._config[config_key]}')
            parameters_map = {}
        return parameters_map

    # fetch the set of rules that will be included/excluded from the CIS rules
    def _get_filter_rules(self, config_key: str, file_key: str) -> List[str]:
        """Get filter rules."""
        try:
            fp = pathlib.Path(self._config[config_key])
            f = fp.open('r', encoding=const.FILE_ENCODING)
            jdata = json.load(f)
            try:
                filter_rules = jdata[file_key]
            except Exception:
                filter_rules = jdata
            f.close()
        except KeyError as e:
            logger.debug(f'key {e.args[0]} missing')
            filter_rules = []
        except Exception:
            logger.warning(f'unable to process {self._config[config_key]}')
            filter_rules = []
        return filter_rules

    # create map from file:
    # key is rule
    # value is tuple comprising [ category, control, description ]
    def _get_cis_rules(self, filename: str) -> Dict[str, Tuple[str, str, str]]:
        """Get CIS rules."""
        try:
            fp = pathlib.Path(filename)
            f = fp.open('r', encoding=const.FILE_ENCODING)
            content = f.readlines()
            rules = self._parse_cis_rules(content)
            f.close()
        except Exception:
            logger.warning(f'unable to process {filename}')
            rules = {}
        return rules

    def _parse_cis_rules(self, content: List[str]) -> Dict[str, Tuple[str, str, str]]:
        """Parse CIS rules."""
        rules = {}
        lineno = 0
        for line in content:
            lineno += 1
            line = line.replace('\n', '')
            if line.startswith('  #### '):
                category = line.split('  #### ')[1]
                logger.debug(f'{lineno} category: {category}')
            elif line.startswith('  # '):
                text = line.split('  # ')[1].split(' ', 1)
                if '.' in text[0]:
                    control = text[0]
                    desc = text[1]
                    logger.debug(f'{lineno} control: {control} description: {desc}')
                else:
                    logger.debug(f'{lineno} skip: {line}')
            elif line.startswith('    - '):
                rule = line.split('    - ')[1]
                logger.debug(f'{lineno} rule: {rule}')
                if not self._is_selected(rule):
                    logger.debug(f'not selected rule: {rule}')
                elif rule in rules.keys():
                    logger.info(f'duplicate rule: {rule}')
                else:
                    rules[rule] = [category, control, desc]
            else:
                logger.debug(f'{lineno} skip: {line}')
        return rules

    # rule is selected if:
    # a) the selected rules file is not specified or is empty or
    # b) the rule appears in the list of selected rules from the file
    def _is_selected(self, rule: str) -> bool:
        """Check if rule is selected."""
        retval = True
        if len(self._selected_rules) > 0 and rule not in self._selected_rules:
            retval = False
        logger.debug(f'{retval} {rule}')
        return retval

    # rule is excluded if it does not appear in the list of trimmed rules
    # for the control
    def _is_excluded(self, rule: str, control: str, controls: Dict[str, List[str]]) -> bool:
        """Check if rule is excluded."""
        retval = False
        if rule not in controls[control]:
            logger.debug(f'exclude {rule} {control}')
            retval = True
        return retval

    def _build_roles(self) -> List[Role]:
        """Build roles."""
        value = [
            Role(id='prepared-by', title='Indicates the organization that created this content.'),
            Role(id='prepared-for', title='Indicates the organization for which this content was created..'),
            Role(
                id='content-approver',
                title='Indicates the organization responsible for all content represented in the "document".'
            ),
        ]
        return value

    def _build_control_implementation(
        self, profile_set: Dict[str, str], responsible_roles: List[ResponsibleRole], props: List[Property]
    ) -> ControlImplementation:
        """Build control implementation."""
        implemented_requirements = self._build_implemented_requirements(profile_set, responsible_roles)
        if len(implemented_requirements) == 0:
            control_implementation = None
        else:
            control_implementation = ControlImplementation(
                uuid=str(uuid.uuid4()),
                source=profile_set['profile-url'],
                description=f'{profile_set["component-name"]} implemented controls for {profile_set["profile-title"]}.',
                implemented_requirements=implemented_requirements,
                props=props,
            )
        return control_implementation

    def _get_title(self, dir_name: str, root: str) -> str:
        """Extract rule title from compliance-as-code rule.yml."""
        """
        Operation:
        Given is a dir_name and a root directory. We walk the
        directory tree looking for a directory named dir_name.
        Once found, we read the content of the rule.yml file in
        that directory. It is likely that we read each rule.yml
        file exactly once, since each rule appears exactly once
        in the one or more profiles, e.g. cis-node.profile, which
        drive the search. From the content, we find the title and
        return its corresponding value.
        """

        title = None
        for path, dirs, _files in os.walk(root):
            if dir_name in dirs:
                folder = os.path.join(path, dir_name)
                tpath = pathlib.Path(folder) / 'rule.yml'
                fp = pathlib.Path(tpath)
                f = fp.open('r', encoding=const.FILE_ENCODING)
                content = f.readlines()
                f.close()
                for line in content:
                    if line.startswith('title:'):
                        title = line.split('title:')[1]
                        break
        if title is None:
            msg = f'unable to find "{dir_name}"'
            logger.warning(msg)
            title = 'no title'
        title = title.strip().strip("'").strip('"')
        logger.debug(f'{title}')
        return title

    def _build_implemented_requirements(self, profile_set: Dict[str, str],
                                        responsible_roles: List[ResponsibleRole]) -> List[ImplementedRequirement]:
        """Build implemented requirements."""
        implemented_requirements = []
        profile_file = profile_set['profile-file']
        rules = self._get_cis_rules(profile_file)
        controls = self._get_controls(rules)
        rule_prefix = 'xccdf_org.ssgproject.content_rule_'
        cac_openshift = f'{self._folder_cac}/applications/openshift'
        for rule in rules:
            if self._is_excluded(rule, rules[rule][1], controls):
                continue
            remarks = self._get_title(rule, cac_openshift)
            prop = Property(
                class_='scc_goal_name_id',
                ns=profile_set['profile-ns'],
                name='XCCDF_rule',
                value=f'{rule_prefix}{rule}',
                remarks=f'{remarks}'
            )
            props = [prop]
            implemented_requirement = ImplementedRequirement(
                uuid=f'{str(uuid.uuid4())}',
                control_id=f'CIS-{rules[rule][1]}',
                description=f'{rules[rule][2]}',
                props=props,
                responsible_roles=responsible_roles,
            )
            set_parameter = self._get_set_parameter(rule)
            if set_parameter is not None:
                implemented_requirement.set_parameters = [set_parameter]
            implemented_requirements.append(implemented_requirement)
        return implemented_requirements

    def _build_responsible_roles(self, party_uuid_01: str, party_uuid_02: str,
                                 party_uuid_03: str) -> List[ResponsibleRole]:
        """Build responsible roles."""
        role_prepared_by = ResponsibleRole(role_id='prepared-by', party_uuids=[party_uuid_01])
        role_prepared_for = ResponsibleRole(role_id='prepared-for', party_uuids=[party_uuid_02, party_uuid_03])
        role_content_approver = ResponsibleRole(role_id='content-approver', party_uuids=[party_uuid_01])
        value = [
            role_prepared_by,
            role_prepared_for,
            role_content_approver,
        ]
        return value

    def _build_parties(
        self, org_name: str, org_remarks: str, party_uuid_01: str, party_uuid_02: str, party_uuid_03: str
    ) -> List[Party]:
        """Build parties."""
        value = [
            Party(uuid=party_uuid_01, type='organization', name=org_name, remarks=org_remarks),
            Party(
                uuid=party_uuid_02,
                type='organization',
                name='Customer',
                remarks='organization to be customized at account creation only for their Component Definition'
            ),
            Party(
                uuid=party_uuid_03,
                type='organization',
                name='ISV',
                remarks='organization to be customized at ISV subscription only for their Component Definition'
            ),
        ]
        return value

    def _build_responsible_parties(self, party_uuid_01: str, party_uuid_02: str,
                                   party_uuid_03: str) -> List[ResponsibleParty]:
        """Build responsible parties."""
        prepared_by = ResponsibleParty(role_id='prepared-by', party_uuids=[party_uuid_01])
        prepared_for = ResponsibleParty(role_id='prepared-for', party_uuids=[party_uuid_02, party_uuid_03])
        content_approver = ResponsibleParty(role_id='content-approver', party_uuids=[party_uuid_01])
        value = [
            prepared_by,
            prepared_for,
            content_approver,
        ]
        return value



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for cli module command version."""
import pathlib

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils
from tests.test_utils import execute_command_and_assert

from trestle import __version__
from trestle.oscal import OSCAL_VERSION


def test_version(monkeypatch: MonkeyPatch) -> None:
    """Test version output."""
    testcmd = 'trestle version'
    execute_command_and_assert(testcmd, 0, monkeypatch)


def test_oscal_obj_version(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch, capsys) -> None:
    """Test OSCAL object version output."""
    test_utils.setup_for_multi_profile(tmp_trestle_dir, False, True)
    testcmd = 'trestle version -n test_profile_c -t profile'
    execute_command_and_assert(testcmd, 0, monkeypatch)
    output, _ = capsys.readouterr()
    assert 'Version of OSCAL object of test_profile_c profile is: 2021-01-01' in output

    testcmd = 'trestle version -n test_profile_c'
    execute_command_and_assert(testcmd, 1, monkeypatch)

    testcmd = 'trestle version -n test_profile_c -t nonexisting'
    execute_command_and_assert(testcmd, 1, monkeypatch)

    testcmd = 'trestle version'
    execute_command_and_assert(testcmd, 0, monkeypatch)
    output, _ = capsys.readouterr()
    assert f'Trestle version v{__version__} based on OSCAL version {OSCAL_VERSION}' in output

    testcmd = 'trestle version -n complex_cat -t catalog'
    execute_command_and_assert(testcmd, 0, monkeypatch)
    output, _ = capsys.readouterr()
    assert 'Version of OSCAL object of complex_cat catalog is: REPLACE_ME' in output

    comp_name = test_utils.setup_for_component_definition(tmp_trestle_dir, monkeypatch)
    testcmd = f'trestle version -n {comp_name} -t component-definition'
    execute_command_and_assert(testcmd, 0, monkeypatch)
    output, _ = capsys.readouterr()
    assert f'Version of OSCAL object of {comp_name} component-definition is: 0.21.0' in output



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Remove Command."""
import argparse
import logging
import pathlib
from typing import List, Tuple

import trestle.common.const as const
import trestle.common.err as err
from trestle.common import log
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.command_docs import CommandPlusDocs
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.models.actions import CreatePathAction, RemoveAction, WriteFileAction
from trestle.core.models.elements import Element, ElementPath
from trestle.core.models.file_content_type import FileContentType
from trestle.core.models.plans import Plan

logger = logging.getLogger(__name__)


class RemoveCmd(CommandPlusDocs):
    """Remove a subcomponent from an existing model."""

    name = 'remove'

    def _init_arguments(self) -> None:
        self.add_argument(
            f'-{const.ARG_FILE_SHORT}',
            f'--{const.ARG_FILE}',
            help=const.ARG_DESC_FILE + ' to remove component/subcomponent to.',
            required=True
        )
        self.add_argument(
            f'-{const.ARG_ELEMENT_SHORT}',
            f'--{const.ARG_ELEMENT}',
            help=const.ARG_DESC_ELEMENT + ' to remove.',
            required=True
        )

    def _run(self, args: argparse.Namespace) -> int:
        """Remove an OSCAL component/subcomponent to the specified component.

        This method takes input a filename and a list of comma-seperated element path. Element paths are field aliases.
        The method first finds the parent model from the file and loads the file into the model.
        Then the method executes 'remove' for each of the element paths specified.
        """
        try:
            log.set_log_level_from_args(args)
            args_dict = args.__dict__

            file_path = pathlib.Path(args_dict[const.ARG_FILE]).resolve()
            relative_path = file_path.relative_to(args.trestle_root)
            # Get parent model and then load json into parent model
            parent_model, parent_alias = ModelUtils.get_relative_model_type(relative_path)

            parent_object = parent_model.oscal_read(file_path)
            parent_element = Element(parent_object, parent_alias)

            add_plan = Plan()

            # Do _remove for each element_path specified in args
            element_paths: List[str] = str(args_dict[const.ARG_ELEMENT]).split(',')
            for elm_path_str in element_paths:
                element_path = ElementPath(elm_path_str)
                remove_action, parent_element = self.remove(element_path, parent_element)
                add_plan.add_action(remove_action)

            create_action = CreatePathAction(file_path, True)
            write_action = WriteFileAction(file_path, parent_element, FileContentType.to_content_type(file_path.suffix))
            add_plan.add_action(remove_action)
            add_plan.add_action(create_action)
            add_plan.add_action(write_action)

            add_plan.execute()

            return CmdReturnCodes.SUCCESS.value

        except Exception as e:
            return err.handle_generic_command_exception(e, logger, 'Error while removing OSCAL component')

    @classmethod
    def remove(cls, element_path: ElementPath, parent_element: Element) -> Tuple[RemoveAction, Element]:
        """For the element_path, remove a model from the parent_element of a given parent_model.

        First we check if there is an existing element at that path
        If not, we complain.
        Then we set up an action plan to update the model (specified by file_path) in memory,
        return the action and return the parent_element.

        LIMITATIONS:
        1. This does not remove elements of a list or dict. Instead, the entire list or dict is removed.
        2. This cannot remove arbitrarily named elements that are not specified in the schema.
        For example, "responsible-parties" contains named elements, e.g., "organisation". The tool will not
        remove the "organisation" as it is not in the schema, but one can remove its elements, e.g., "party-uuids".
        """
        element_path_list = element_path.get_full_path_parts()
        if '*' in element_path_list:
            raise err.TrestleError('trestle remove does not support Wildcard element path.')

        deleting_element = parent_element.get_at(element_path)

        if deleting_element is not None:
            # The element already exists
            if type(deleting_element) is list:
                logger.warning(
                    'Warning: trestle remove does not support removing elements of a list: '
                    'this removes the entire list'
                )
            elif type(deleting_element) is dict:
                logger.warning(
                    'Warning: trestle remove does not support removing dict elements: '
                    'this removes the entire dict element'
                )
        else:
            raise err.TrestleError(f'Bad element path: {str(element_path)}')

        remove_action = RemoveAction(parent_element, element_path)

        return remove_action, parent_element
